import os,sys,json,copy,traceback,glob
from . import comm_echarts,comm_func
from dp.launching.report import Report, AutoReportElement, ReportSection, ChartReportElement

# the unit of some metrics
METRICS_UNIT = {
    "energy": "eV",
    "energy_per_atom": "eV/atom",
    "force": "eV/A",
    "stress": "kbar",
    "volume": "A^3",
    "efermi": "eV",
    "band_gap": "eV",
    "virial": "eV",
    "pressure": "kbar",
    "total_time": "s",
    "largest_gradient": "eV/A",
    "stress_time": "s",
    "force_time": "s",
    "scf_time": "s",
    "step1_time": "s",
    "lattice_constant": "A",
    "denergy_last": "eV",
    "denergy_womix_last": "eV"
}

def add_unit_metrics(metrics):
    # metrics = {"example1": {"metric1": value1, "metric2": value2, ...}, "example2": {"metric1": value1, "metric2": value2, ...}, ...}
    # modify the name of metrics to metric(unit)
    # return the new metrics
    new_metrics = {}
    for iexample, ivalue in metrics.items():
        new_metrics[iexample] = {}
        for imetric, iv in ivalue.items():
            if imetric in METRICS_UNIT:
                new_metrics[iexample][f"{imetric}({METRICS_UNIT[imetric]})"] = iv
            else:
                new_metrics[iexample][imetric] = iv
    return new_metrics

def add_unit(metric_name):
    # metric_name is a string
    # return the new metric_name with unit
    if metric_name in METRICS_UNIT:
        return f"{metric_name}({METRICS_UNIT[metric_name]})"
    else:
        return metric_name

def sort_lists(lists):
    # lists is a list of list
    # lists = [[x1,x2,...],[y1,y2,...],...]
    # return the sorted lists
    # the first list will be sorted, and the other lists will be sorted according to the first list
    # if the length of lists is not equal, return the original lists
    if len(lists) in [0,1]:
        return lists
    else:
        # check if all elements are list and the length of each list is equal
        if False in [isinstance(i,list) for i in lists]:
            return lists
        if len(set([len(i) for i in lists])) != 1:
            return lists
        
        lists_tmp = copy.deepcopy(lists)
        new_lists = [[] for i in range(len(lists))]
        for ii in sorted(zip(*lists_tmp)):
            for ij,jj in enumerate(list(ii)):
                new_lists[ij].append(jj)
        
        return new_lists

def add_ref(allresults, ref_data):
    # allresults: the metrics data
    # ref_data: the reference data
    # return the metrics data with reference data
    # allresults = {"example1":{"metric1":value1,"metric2":value2,...},"example2":{"metric1":value1,"metric2":value2,...},...}
    # ref_data = {"ref1":{"example1":{"metric1":value1,"metric2":value2,...},"example2":{"metric1":value1,"metric2":value2,...},...},"ref2":{"example1":{"metric1":value1,"metric2":value2,...},"example2":{"metric1":value1,"metric2":value2,...},...},...}
    # 1. the reference metrics wiil be named as "<metrics>_ref_<name of reference>"
    # 2. if example/metrics in ref_data is not in allresults, the ref will not be added
    # 3. if example in ref_data is a father path of example in allresults, the ref will be added

    # get the example name
    example_name = list(allresults.keys())
    
    # get the metrics name
    metric_name = []
    for ivalue in allresults.values():
        metric_name += list(ivalue.keys())
    metric_name = list(set(metric_name))
    
    # get the reference name
    ref_name = list(ref_data.keys())
    # get the metric_name in ref_data
    ref_metric_name = []
    for iref in ref_name:
        for imetric in ref_data[iref].values():
            ref_metric_name += list(imetric.keys())
    #only metrics in metric_name will be added
    #print(ref_name,ref_metric_name,metric_name)
    ref_metric_name = list(set(ref_metric_name) & set(metric_name))
    
    # if ref_metric_name is empty, return
    if len(ref_metric_name) == 0:
        return allresults,[]
    
    new_results = {}
    for ikey,ivalue in allresults.items():
        new_results[ikey] = ivalue.copy()
        
        for iref in ref_name:
            # check if ikey in ref_data[iref]
            if ikey in ref_data[iref]:
                for imetric in ref_metric_name:
                    new_results[ikey][f"{imetric}_ref_{iref}"] = ref_data[iref][ikey].get(imetric,None)
            else:
                # check if path in iref is a father path of ikey
                for iref_example in ref_data[iref]:
                    if ikey.startswith(iref_example) and ikey[len(iref_example)] == "/":
                        for imetric in ref_metric_name:
                            new_results[ikey][f"{imetric}_ref_{iref}"] = ref_data[iref][iref_example].get(imetric,None)
                        break
    return new_results,[f"{imetric}_ref_{iref}" for imetric in ref_metric_name for iref in ref_name]   

def check_example_name(example_name):
    '''
    Check if the example name is a/00000, a/00001, ..., b/00000, b/00001
    If yes will return the prefix and the number list.
    If not will return None and None
    '''
    example_name_prefix = []
    example_name_number = []
    has_prefix = True
    is_number = True
    for i in example_name:
        if not "/" in i:
            has_prefix = False
            break
        tmp = i.split("/")
        if len(tmp) != 2:
            has_prefix = False
            break
        try:
            _ = int(tmp[1])
        except:
            is_number = False
            break
        example_name_prefix.append(tmp[0])
        example_name_number.append(tmp[1])
    if has_prefix and is_number:
        example_name_prefix = list(set(example_name_prefix))
        example_name_number = list(set(example_name_number))
        example_name_prefix.sort()
        example_name_number.sort()
        return example_name_prefix,example_name_number
    else:
        return None,None

def gen_multiple_y(x, ys, legend_list, example_name_prefix, example_name_number, has_yref = True):
    '''
    split x, ys and legend_list to several parts according to example_name_prefix and example_name_number
    if ys has more than 2 lists, the 1st list is the value of this job, and the other lists are the values of different reference
    And, there may has some same value in ys[1:], because for different example_name_number, the reference are same.
    
    if has_yref = False, then all ys are the values that needed to be ploted, and the legend_list is the legend of ys
    the new legend will be the mix of example_name_number and legend_list.
    '''
    if example_name_prefix == None or example_name_number == None:
        return legend_list, x, ys
    
    x_tmp = copy.deepcopy(example_name_prefix)
    y_tmp = []
    for i in range(len(legend_list) * len(example_name_number)):
        y_tmp.append([])
  
    for idx,iexample in enumerate(example_name_prefix):
        for jdx, jnumber in enumerate(example_name_number):
            example_name_complete = iexample + "/" + jnumber
            if example_name_complete in x:
                iidx = x.index(example_name_complete)
                for iy in range(len(ys)):
                    # y_tmp = [legend1_example_name_number1_legend1,legend1_example_name_number2,..., legend2_example_name_number1,... ]
                    y_tmp[iy*len(example_name_number) + jdx].append(ys[iy][iidx])
            else:
                for iy in range(len(ys)):
                    y_tmp[iy*len(example_name_number) + jdx].append(None)
    
    if has_yref:
        legend_tmp = copy.deepcopy(example_name_number )
        if len(legend_list) > 1:
            legend_tmp += legend_list[1:]
        new_ytmp = y_tmp[0:len(example_name_number)]
        # only need legend1_example_name_number1,..,legend1_example_name_numbern, legend2_example_name_number1,lengend3_example_name_number1,...legendn_example_name_number1
        for iy in range(1,len(ys)):
            new_ytmp.append(y_tmp[iy*len(example_name_number)])
        y_tmp = new_ytmp
    else:
        legend_tmp = []
        for i in legend_list:
            for j in example_name_number:
                legend_tmp.append(i + "_" + j)
    #print(len(legend_tmp),len(x_tmp),len(y_tmp),len(ys),len(example_name_number),example_name_number)
    return legend_tmp, x_tmp, y_tmp                
    
def plot_delta_Y(y_list, legend_list, example_name, imetric,example_name_prefix, example_name_number):
    # y_list is a list of list
    # y_list = [[y1,y2,...],[y1,y2,...],...]
    # return the delta y_list and percentage delta y_list
    # the first list is the value of this job, and the other lists are the values of different reference
    # length of y_list should be at least 2, and equal to length of legend_list
    chart_elements = []
    
    def flat_list(list1):
        new_list = []
        for i in list1:
            if isinstance(i,list):
                new_list += flat_list(i)
            else:
                new_list.append(i)
        return new_list
    
    # need to check if y_list[0] is a list of list
    #print(imetric)
    if True in [isinstance(i,list) for i in y_list[0]]:
        # if y_list[0] is a list of list, we need to plot the delta Y
        #print(y_list)
        new_y_list = copy.deepcopy(y_list)
        # need to transfer each list to one dimension list
        for iy in range(len(new_y_list)):
            for iiy in range(len(new_y_list[iy])):
                if isinstance(new_y_list[iy][iiy],list):
                    new_y_list[iy][iiy] = flat_list(new_y_list[iy][iiy])
        
        #print("new_y_list",new_y_list)
        #print(len(new_y_list),len(new_y_list[0]))
        delta_y = []
        for iy in range(1,len(new_y_list)): # loop for each reference
            delta_y.append([])
            for jy in range(len(new_y_list[0])): # loop for each example
                if not isinstance(new_y_list[0][jy],list) or not isinstance(new_y_list[iy][jy],list) or len(new_y_list[0][jy]) != len(new_y_list[iy][jy]):
                    #print(type(new_y_list[0][jy]),type(new_y_list[iy][jy]))
                    delta_y[-1].append(None)
                    continue
                try:
                    ivalue = [new_y_list[0][jy][i] - new_y_list[iy][jy][i] for i in range(len(new_y_list[0][jy]))]
                    delta_y[-1].append(ivalue)
                except:
                    delta_y[-1].append(None)
        # now we need to calculate some statistics
        # 1. the max absolute value
        # 2. the norm of delta_y
        
        max_abs = []
        norm = []
        legend_abs = []
        legend_norm = []
        for iy in range(len(delta_y)):# loop for each reference
            iabs = []
            inorm = []
            for jy in range(len(delta_y[iy])):# loop for each example
                if delta_y[iy][jy] == None:
                    iabs.append(None)
                    inorm.append(None)
                    continue
                try:
                    ivalue = max([abs(i) for i in delta_y[iy][jy]])
                except:
                    ivalue = None
                iabs.append(ivalue)
                try:
                    ivalue = sum([i**2 for i in delta_y[iy][jy]])**0.5/len(delta_y[iy][jy])
                except:
                    ivalue = None
                inorm.append(ivalue)
            if set(iabs) != {None}:
                max_abs.append(iabs)
                legend_abs.append(legend_list[iy+1])
            if set(inorm) != {None}:
                norm.append(inorm)
                legend_norm.append(legend_list[iy+1])
        
        # plot max_abs and norm
        
        if len(max_abs) > 0:
            tmp_ = sort_lists([example_name]+max_abs)
            legend_tmp, x_tmp, y_tmp = gen_multiple_y(tmp_[0], tmp_[1:], legend_abs, example_name_prefix, example_name_number, has_yref = False)
            options = comm_echarts.produce_multiple_y(f"{add_unit(imetric)}(max(abs(this job - reference)))", x_tmp, y_tmp, legend_tmp, x_type="category", y_type="value")
            options["xAxis"][0]["axisLabel"] = {
                "rotate": 15,
                "interval": int(len(example_name)/15)
                } 
            chart_elements.append(ChartReportElement(
                    options=options, title=f"{add_unit(imetric)}(max(abs(this job - reference)))"))
        if len(norm) > 0:
            tmp_ = sort_lists([example_name]+norm)
            legend_tmp, x_tmp, y_tmp = gen_multiple_y(tmp_[0], tmp_[1:], legend_norm, example_name_prefix, example_name_number, has_yref = False)
            options = comm_echarts.produce_multiple_y(f"{add_unit(imetric)}(Norm(this job - reference)/N)", x_tmp, y_tmp, legend_tmp, x_type="category", y_type="value")
            options["xAxis"][0]["axisLabel"] = {
                "rotate": 15,
                "interval": int(len(example_name)/15)
                } 
            chart_elements.append(ChartReportElement(
                    options=options, title=f"{add_unit(imetric)}(Norm(this job - reference)/N)"))      
    else:
        delta_y_list = []
        percentage_delta_y_list = []
        delta_legend_list1 = []
        delta_legend_list2 = []

        for iy in range(1,len(y_list)):
            # need to check if the two value can do minus
            values1 = []
            values2 = []
            for i in range(len(y_list[iy])):
                if isinstance(y_list[0][i],(int,float)) and isinstance(y_list[iy][i],(int,float)):
                    ivalue = y_list[0][i] - y_list[iy][i]                     
                else:
                    ivalue = None
                values1.append(ivalue)

                if isinstance(y_list[0][i],(int,float)) and isinstance(y_list[iy][i],(int,float)) and y_list[iy][i] != 0:
                    ivalue = (y_list[0][i] - y_list[iy][i])/y_list[iy][i]
                else:
                    ivalue = None
                values2.append(ivalue)
            if set(values1) != {None}:
                delta_y_list.append(values1)
                delta_legend_list1.append(legend_list[iy])
            if set(values2) != {None}:
                percentage_delta_y_list.append(values2)
                delta_legend_list2.append(legend_list[iy])
        if len(delta_y_list) > 0:  
            tmp_ = sort_lists([example_name]+delta_y_list)
            legend_tmp, x_tmp, y_tmp = gen_multiple_y(tmp_[0], tmp_[1:], delta_legend_list1, example_name_prefix, example_name_number, has_yref = False)
            options = comm_echarts.produce_multiple_y(f"{add_unit(imetric)}(Delta = this job - reference)", x_tmp, y_tmp, legend_tmp, x_type="category", y_type="value")
            options["xAxis"][0]["axisLabel"] = {
                "rotate": 15,
                "interval": int(len(example_name)/15)
                } 
            chart_elements.append(ChartReportElement(
                    options=options, title=f"{add_unit(imetric)}(Delta = this job - reference)"))
        if len(percentage_delta_y_list) > 0:
            tmp_ = sort_lists([example_name]+percentage_delta_y_list)
            legend_tmp, x_tmp, y_tmp = gen_multiple_y(tmp_[0], tmp_[1:], delta_legend_list2, example_name_prefix, example_name_number, has_yref = False)
            options = comm_echarts.produce_multiple_y(f"{add_unit(imetric)}(Delta/Reference)", x_tmp, y_tmp, legend_tmp, x_type="category", y_type="value")
            options["xAxis"][0]["axisLabel"] = {
                "rotate": 15,
                "interval": int(len(example_name)/15)
                } 
            chart_elements.append(ChartReportElement(
                    options=options, title=f"{add_unit(imetric)}(Delta/Reference)"))
    return chart_elements

def plot_two_metrics(pddata,
                     x_name, y_name,
                     example_name,
                     ref_metric_name=[],
                     x_type="category", y_type="value", shift_type=0):
    '''
    pddata: the pandas data
    ref_metric_name: the reference metric name
    x_name: the name of x
    y_name: the name of y
    example_name: a list of example name
    x_type: the type of x in echart  # now only support category
    y_type: the type of y in echart
    shift_type: the type to shift the y value
        -1: has reference, will minus the reference value
        0: do not shift
        1: shift to make the min value is 0
        2: shift to make the last value is 0
    if has reference, all value will minus the reference value
    if has multiple reference, will minus each reference, and set shift_type to -1
    '''
    x_type = "category"

    if x_name not in pddata.index.to_list() or y_name not in pddata.index.to_list():
        return []

    # check if has reference, only chech for y_name
    ref_names = []
    ref_values = []
    for imetric in ref_metric_name:
        if imetric.startswith(f"{y_name}_ref_") and imetric in pddata.index.to_list():
            ref_names.append(imetric.split("_ref_")[-1])
            ref_values.append(pddata.loc[imetric, :].to_list())
            shift_type = -1
    
    #print("x_name,y_name:", x_name, y_name)
    all_x = pddata.loc[x_name, :].to_list()
    all_y = pddata.loc[y_name, :].to_list()
    chart_elements = []
    if len(set(all_x)) <= 1:
        return []

    # the example name may be a/00000, a/00001, ..., b/00000, b/00001
    # we need plot each chart for a, b, ...
    # so we need to split the example name to get the prefix
    # and then plot the chart for each prefix
    # we need to plot the chart for all example that do not match the prefix/00000 format
    # we will seperate the example to several parts: prefix/00000 and others
    all_xy = {"": [[], [], f"{x_name} VS {y_name}"]}
    for i in range(len(ref_names)): 
        all_xy[""].append([])  # add a new list for each reference

    # remove / in the end of the example_name
    example_name = [i.rstrip("/") for i in example_name]
    for i, ie in enumerate(example_name):
        prefix = os.path.dirname(ie)
        basename = os.path.basename(ie)
        if basename.isdigit():
            if prefix not in all_xy:
                all_xy[prefix] = [[], [], f"{x_name} vs {y_name} ({prefix})"]
                for iref in range(len(ref_names)):
                    all_xy[prefix].append([])
            all_xy[prefix][0].append(all_x[i])
            all_xy[prefix][1].append(all_y[i])
            for iref in range(len(ref_names)):
                all_xy[prefix][iref+3].append(ref_values[iref][i])
        else:
            all_xy[""][0].append(f"{all_x[i]}({prefix})")
            all_xy[""][1].append(all_y[i])
            for iref in range(len(ref_names)):
                all_xy[""][iref+3].append(ref_values[iref][i])

    # if one prefix only has one example, add the example to ""
    for iprefix, ivalue in all_xy.items():
        if len(ivalue[0]) == 1:
            all_xy[""][0].append(f"{ivalue[0][0]}({iprefix})")
            all_xy[""][1].append(ivalue[1][0])
            for iref in range(len(ref_names)):
                all_xy[""][iref+3].append(ivalue[iref+3][0])
            del all_xy[iprefix]

    # plot the chart for each prefix
    for iprefix, ivalue in all_xy.items():
        # if ivalues is empty or only one value, continue
        if len(ivalue[0]) <= 1:
            continue

        x = ivalue[0]
        y = [ivalue[1]]
        for iref in range(len(ref_names)):
            y.append(ivalue[iref+3])
        title_full = title = add_unit(ivalue[2])

        # need shift the y
        y_real = [i for i in y[0] if i != None]  # get the none-none value of y[0]
        if len(y_real) == 0:
            continue
        legend = [y_name]
        if shift_type == 1:
            title_full = title_full + " (minus the minimum)"
            min_e = min(y_real)
            y = [[i if i == None else i-min_e for i in y[0]]]
        elif shift_type == 2:
            title_full = title_full + " (minus the last value)"
            min_e = y_real[-1]
            y = [[i if i == None else i-min_e for i in y[0]]]
        elif shift_type == -1:
            title_full = title_full + f" (minus reference)"
            title += f" (minus reference)"
            new_y = [[] for i in range(len(y)-1)]
            for iy in range(len(y[0])):
                for iy2 in range(len(y)-1):
                    if y[0][iy] != None and y[iy2+1][iy] != None:
                        new_y[iy2].append(y[0][iy]-y[iy2+1][iy])
                    else:
                        new_y[iy2].append(None) 
            y = new_y
            legend = [f"ref_{i}" for i in ref_names]
        if y_type == "log":
            title_full = title_full + " (abs(delta_y))"
        
        tmp_ = sort_lists([x]+y)     
        options =comm_echarts.produce_multiple_y(
            title, tmp_[0], tmp_[1:], legend, x_type=x_type, y_type=y_type)
        
        chart_elements.append(ChartReportElement(options=options, title=title_full))
    return chart_elements


def plot_trend(data_input, example_input,data_name,logy = True):
    # data_input: a list of data of all examples
    # example_input: a list of example name
    # return a list of ChartReportElement

    chart_report = []
    # only plot data if data is not None
    data_list = []
    example_name = []
    for i in range(len(data_input)):
        if isinstance(data_input[i], list):
            data_list.append(data_input[i])
            example_name.append(example_input[i])
    if len(data_list) == 0:
        return []

    # sort the example_name and data_list
    example_name, data_list = zip(*sorted(zip(example_name, data_list)))
    max_example_in_one_chart = 10
    all_data = []
    for i in range(0, len(example_name), max_example_in_one_chart):
        end = i+max_example_in_one_chart if i + \
            max_example_in_one_chart < len(example_name) else len(example_name)
        all_data.append([example_name[i:end], data_list[i:end]])
    idata = 0
    for legend, data in all_data:
        x = [i+1 for i in range(max([len(j) for j in data]))]
        options = comm_echarts.produce_multiple_y(
            f"{data_name}{idata}", x, data, legend, x_type="category", y_type="log" if logy else "value")
        chart_report.append(ChartReportElement(
            options=options, title=f"{data_name}{idata}"))
        idata += 1
    return chart_report

def fill_none(allresults):
    # if some metrics is not in some example, we need to add None to the example
    allkeys = []
    for ivalue in allresults.values():
        allkeys += list(ivalue.keys())
    allkeys = list(set(allkeys))
    for ikey in allkeys:
        for iexample in allresults:
            if ikey not in allresults[iexample]:
                allresults[iexample][ikey] = None

def produce_metrics(metric_file, output_path, ref_data={}, report_titile="metrics"):
    from abacustest import outresult
    import pandas as pd

    metric_filename = os.path.split(metric_file)[-1]
    allresults = json.load(open(metric_file))
    allresults,ref_metric_name = add_ref(allresults, ref_data)
    csv_filename = os.path.splitext(metric_filename)[0] + ".csv"
    _, _, savefile_names = outresult.pandas_out(
        add_unit_metrics(allresults), os.path.join(output_path, csv_filename),print_result=False)
    report_elements = []
    for ifilename in savefile_names:
        ifilename = os.path.split(ifilename)[-1]
        report_elements.append(AutoReportElement(
            title=os.path.splitext(ifilename)[0], path=ifilename, description=""))

    chart_elements = []
    # produce the Echarts option for each metric
    fill_none(allresults)
    pddata = pd.DataFrame.from_dict(allresults)
    example_name = pddata.columns.to_list()  # get the column name
    metric_name = pddata.index.to_list()  # get the row/index name
    type_set = (int, float, type(None), bool,list)
    
    # check if the example_name has more than 2 levels, and is end with number
    example_name_prefix, example_name_number = check_example_name(example_name)
    print("ref_metric_name",ref_metric_name)
    for imetric in metric_name:
        if imetric in ref_metric_name:
            continue
        ivalue = pddata.loc[imetric, :].to_list()
        y_type = "value"
        if imetric in ["drho_last", "denergy_last", "denergy_womix_last"]:
            y_type = "log"
        if False not in [isinstance(i, type_set) for i in ivalue]:
            #check if imetric has refence
            ref_type = []
            for iref_metric in ref_metric_name:
                if iref_metric.startswith(f"{imetric}_ref_"):
                    ref_type.append(iref_metric.split("_ref_")[-1])
                    
            # if imetric has refence, we need to plot the imetric and its reference together
            # we need to split the imetric to imetric and its reference
            # comm_echarts.produce_multiple_y(produce_multiple_y(title,x,y_list,legend_list,x_type="category",y_type="value")
            y_list = [ivalue]
            legend_list = ["This_Job"]
            for iref in ref_type:
                if f"{imetric}_ref_{iref}" in metric_name:
                    y_list.append(pddata.loc[f"{imetric}_ref_{iref}", :].to_list())
                    legend_list.append(f"ref_{iref}")
                    
            if True not in [isinstance(i,list) for i in ivalue]:
                # do not plot a list
                #print(y_list,legend_list)
                tmp_ = sort_lists([example_name]+y_list)
                
                # if the example_name has more than 2 levels, and is end with number
                # we need to plot the chart with multiple Y with each number as a legend
                legend_tmp, x_tmp, y_tmp = legend_list, tmp_[0], tmp_[1:]
                legend_tmps = [legend_tmp]
                x_tmps = [x_tmp]
                y_tmps = [y_tmp]
                if example_name_prefix != None and example_name_number != None:
                    legend_tmp, x_tmp, y_tmp = gen_multiple_y(tmp_[0], tmp_[1:], legend_list, example_name_prefix, example_name_number)
                    legend_tmps.append(legend_tmp)
                    x_tmps.append(x_tmp)
                    y_tmps.append(y_tmp)
                for legend_tmp, x_tmp, y_tmp in zip(legend_tmps, x_tmps, y_tmps):
                    try:    
                        options = comm_echarts.produce_multiple_y(add_unit(imetric), x_tmp, y_tmp, legend_tmp, x_type="category", y_type=y_type)
                        options["xAxis"][0]["axisLabel"] = {
                                "rotate": 15,
                                "interval": int(len(x_tmp)/15)
                            } 
                        chart_elements.append(ChartReportElement(
                                options=options, title=add_unit(imetric)))
                    except:
                        traceback.print_exc()
                        print("Error: produce multiple y failed!")
            
            if len(ref_type) > 0:
                # plot the delta Y and percentage delta Y
                try:
                    delta_y_elements = plot_delta_Y(y_list, legend_list, example_name, imetric,example_name_prefix, example_name_number )
                    chart_elements += delta_y_elements
                except:
                    traceback.print_exc()
                    print("Error: plot delta Y failed!")
                    

    # produce some special case chart
    # 1. if metrics has ecutwfc/kspacing and energy_per_atom, produce the ecutwfc vs energy_per_atom chart
    print(f"plot {metric_file} metric name:", metric_name)
    for x_name, y_name, y_type, shift_type in [
        ["INPUT/ecutwfc", "energy_per_atom","value",1],
        ["INPUT/kspacing", "energy_per_atom","value",1],
        ["INPUT/lcao_ecut", "energy_per_atom","log",2],
        ["INPUT/lcao_ecut", "band_gap","value",0],
        ["INPUT/smearing_sigma", "energy_per_atom","value",1],
        ]:
        '''
        shift_type: the type to shift the y value
        0: do not shift
        1: shift to make the min value is 0
        2: shift to make the last value is 0
        '''
        try:
            chart_elements += plot_two_metrics(pddata,
                                               x_name, y_name,
                                               example_name,
                                               ref_metric_name,
                                               x_type="category", y_type=y_type, shift_type=shift_type)
        except:
            traceback.print_exc()
            print("Error: plot two metrics failed!", x_name, y_name)

    # plot the trend of drho, denergy, denergy_womix
    for trend_chart in ["drho", "denergy", "denergy_womix"]:
        if trend_chart in metric_name:
            try:
                data_list = pddata.loc[trend_chart, :].to_list()
                chart_elements += plot_trend(data_list, example_name, trend_chart)
            except:
                traceback.print_exc()
                print(f"Error: plot {trend_chart} failed!")

    return report_elements, chart_elements


def produce_supermetrics(supermetric_file, output_path, work_path, save_path, report_titile="supermetrics"):
    import pandas as pd

    supermetrics_filename = os.path.split(supermetric_file)[-1]
    super_metrics = json.load(open(supermetric_file))
    normal_metrics = {}
    special_metrics = {}
    for ikey, ivalue in super_metrics.items():
        if isinstance(ivalue, dict):
            special_metrics[ikey] = ivalue
        else:
            if isinstance(ivalue,list) and len(ivalue) == 1:
                normal_metrics[ikey] = ivalue[0]
            else:
                normal_metrics[ikey] = ivalue

    report_element = None
    if normal_metrics:
        pddata = pd.DataFrame(normal_metrics, index=[0],)
        t_pddata = pddata.transpose()
        print(t_pddata)
        csv_filename = os.path.splitext(supermetrics_filename)[0] + ".csv"
        t_pddata.to_csv(os.path.join(output_path, csv_filename), index=True, header=True)
        report_element = AutoReportElement(
            title=report_titile, path=csv_filename, description="")

    special_section = None
    special_elements = []
    if special_metrics:
        for ikey, ivalue in special_metrics.items():
            file_name = ivalue.get("file", None)
            if file_name:
                ifile_name = os.path.join(work_path, save_path, file_name)
                if os.path.isfile(ifile_name):
                    special_elements.append(AutoReportElement(
                        title=ikey, path=os.path.join(save_path, file_name), description=""))

    if special_elements:
        special_section = ReportSection(
            title=report_titile, elements=special_elements)

    return report_element, special_section, normal_metrics

def gen_sm_tag(allparams):
    # generate the tag for supermetrics to save on launching
    # the tag include below information:
    # 1. lbg_username from allparams["config"]
    # 2. schedule name from allparams["config"]["dflow_labels"]["launching-schedule"]
    # 3. job name from allparams["config"]["dflow_labels"]["launching-job"]
    return [
        f"user_{allparams.get('config',{}).get('bohrium_username','')}",
        f"schedule_{allparams.get('config',{}).get('dflow_labels',{}).get('launching-schedule','')}",
        f"job_{allparams.get('config',{}).get('dflow_labels',{}).get('launching-job','')}"    
    ]

def judge_sm(x,criteria):
    if criteria == None:
        return None
    if isinstance(criteria,str):
        try:
            print("SuperMetric value:",x, "\ncriteria:",criteria)
            sm_pass = eval(criteria)
            return bool(sm_pass)
        except:
            print("Error: the supermetrics criteria string is not correct!")
            return False
    else:
        print("Error: the supermetrics  criteria string is not correct!")
        return False
            

def produce_metrics_superMetrics_reports(allparams, work_path, output_path):

    save_path = allparams.get("save_path", "results")
    reports = []
    allmetrics_files = []
    allsupermetrics_files = []

    # produce the metrics reports 
    metrics_block = allparams.get("post_dft", {}).get("metrics", [])
    if not isinstance(metrics_block, list):
        metrics_block = [metrics_block]
    for imetrics_block in metrics_block:
        if isinstance(imetrics_block,dict):
            # 1. metrics from the custome defined metrics file
            if imetrics_block.get("value_from_file",None):
                allmetrics_files.append(os.path.join(
                    work_path, save_path, str(imetrics_block.get("value_from_file"))))
                
            # 2. metrics from the metrics file
            if imetrics_block.get("save_file",None):
                allmetrics_files.append(os.path.join(
                    work_path, save_path, str(imetrics_block.get("save_file"))))

    # 3. metrics from the undefined metrics file
    allmetrics_files += glob.glob(os.path.join(work_path,
                                  save_path, "metric*.json"))
    allmetrics_files = list(set(allmetrics_files))
    # 4. support the compare with reference, the reference file name shuold be "metrics_ref.json"
    # we need to find the reference file and get the reference metrics.
    # the format of the reference file is the same as the metrics file
    # or the key is the name of reference, and the value is the metrics
    # the reference metrics wiil be named as "<metrics>_ref_<name of reference>"
    # we first read the reference file and get the reference metrics, and save to a dict
    # if the ref file is metrics format, then the refence name will be ""
    ref_data = {}
    ref_file = os.path.join(work_path, save_path, "metrics_ref.json")
    if os.path.isfile(ref_file):
        ref_metrics = json.load(open(ref_file))
        # need to check if the reference metrics is a dict of dict of dict
        format_ok = True
        Three_layer = True
        if isinstance(ref_metrics, dict):
            for ikey, ivalue in ref_metrics.items():
                if not isinstance(ivalue, dict):
                    format_ok = False
                    break
                if Three_layer:
                    for jkey, jvalue in ivalue.items():
                        if not isinstance(jvalue, dict):
                            Three_layer = False
                            break
        else:
            format_ok = False
        if format_ok:
            if Three_layer:
                ref_data = ref_metrics
            else:
                ref_data = {"": ref_metrics}
                
        if ref_file in allmetrics_files:
            allmetrics_files.remove(ref_file)
        
    metrics_report = []
    metrics_chart_elements = []
    for metric_file in list(set(allmetrics_files)):
        print("PRODUCE REPORT for metric file:", metric_file)
        if not os.path.isfile(metric_file):
            print(f"\tError: metric file \"{metric_file}\" does not exist! Skip it!")
            continue
        try:
            metric_filename = os.path.split(metric_file)[-1]
            tmp_report_elements, tmp_chart_elements = produce_metrics(
                metric_file, output_path, ref_data=ref_data,report_titile=metric_filename)
            if tmp_report_elements:
                metrics_report += tmp_report_elements
            if tmp_chart_elements:
                metrics_chart_elements += tmp_chart_elements
        except:
            traceback.print_exc()
            print(f"Error: report metrics from file \"{metric_file}\" failed!")

    # produce the supermetrics reports
    # 1. supermetrics from the custome defined supermetrics file
    super_metric_filename = allparams.get("post_dft", {}).get(
        "super_metrics", [{}])[0].get("save_file")
    if super_metric_filename:
        allsupermetrics_files.append(os.path.join(
            work_path, save_path, super_metric_filename))

    # 2. supermetrics from the supermetrics file
    customized_super_metrics_filename = allparams.get("post_dft", {}).get(
        "super_metrics", [{}])[0].get("value_from_file", None)
    if customized_super_metrics_filename:
        allsupermetrics_files.append(os.path.join(
            work_path, save_path, customized_super_metrics_filename))

    # 3. supermetrics from the undefined supermetrics file
    allsupermetrics_files += glob.glob(os.path.join(work_path, save_path, "superMetric*.json")) + \
        glob.glob(os.path.join(work_path, save_path, "super_metric*.json")) + \
        glob.glob(os.path.join(work_path, save_path, "supermetric*.json"))
    
    # 4. read the criteria.json, which defined criteria for supermetrics
    if os.path.isfile(os.path.join(work_path, save_path, "criteria.json")):
        criterias = json.load(open(os.path.join(work_path, save_path, "criteria.json")))
    else:
        criterias = {}
    print("criteria:",criterias)
    
    supermetrics_report = []
    supermetrics_special_section = []
    supermetrics_save = {}
    # the summary of supermetrics, this table will be tranlated to a html table
    sm_summary = [["Super Metric","Value","Criteria"]]   
    for super_metric_file in list(set(allsupermetrics_files)):
        print("PRODUCE REPORT for supermetric file:", super_metric_file)
        if not os.path.isfile(super_metric_file):
            print(
                f"\tError: supermetric file \"{super_metric_file}\" does not exist! Skip it!")
            continue
        try:
            super_metric_filename = os.path.split(super_metric_file)[-1]
            tmp_report_element, tmp_special_section, tmp_smetrics = produce_supermetrics(
                super_metric_file, output_path, work_path, save_path, report_titile=super_metric_filename)
            if tmp_report_element:
                supermetrics_report.append(tmp_report_element)
            if tmp_special_section:
                supermetrics_special_section.append(tmp_special_section)
            #print("tmp_smetrics:",tmp_smetrics)
            if tmp_smetrics:
                failcount = 0
                for ik,iv in tmp_smetrics.items():
                    sm_summary.append([ik,iv,criterias.get(ik,None)])
                    cri_tmp = criterias.get(ik,None)
                    sm_pass = judge_sm(iv,cri_tmp)
                    if sm_pass == True:
                        # if True, set the iv to green
                        sm_summary[-1][1] = "<font color=\"green\">"+str(iv)+"</font>"
                    elif sm_pass == False:
                        # if False, set the iv to red
                        sm_summary[-1][1] = "<font color=\"red\">"+str(iv)+"</font>"
                        failcount += 1
                            
                    if not isinstance(iv,(int,float,bool)):
                        continue
                    sname = ik
                    if ik in supermetrics_save:
                        print(f"\tWarning: supermetric \"{ik}\" is already in the supermetrics_save, will be replaced by the new one!")
                        sname = ik + "_" + os.path.splitext(os.path.basename(super_metric_file))[0]
                    if isinstance(iv,bool):
                        supermetrics_save[sname+":int"] = 1 if iv else 0
                    else:
                        supermetrics_save[sname+":float"] = iv
                if supermetrics_save:
                    supermetrics_save["abacustest_TEST_PASS_ratio:float"] = 1.0 - float(failcount)/len(tmp_smetrics)
                    print(supermetrics_save)
        except:
            traceback.print_exc()
            print(
                f"Error: report supermetrics from file \"{super_metric_file}\" failed!")

    # produce the summary of supermetrics
    if len(sm_summary) > 1:
        sm_summary = comm_func.produce_html_table(sm_summary)
        criterias_comment = criterias.get("__COMMENT","")
        # write the summary to a html file
        with open(os.path.join(output_path,"supermetrics_summary.html"),"w") as f:
            # firstly add the head of html
            # set the table width to 100%, and set content to center
            # add a title to the table, "supermetrics summary"
            # add the criterias_comment to the end of the table
            f.write("<html><head><meta charset=\"utf-8\"><style>")
            f.write("table {width:85%;text-align:center;}")
            f.write(".note {width: 75%;text-align: left;display: inline-block;word-wrap: break-word;white-space: pre-wrap;}")
            f.write("</style></head><body>")
            f.write("<table border=\"2px\"><thead><tr><td>Super Metrics Summary</td></tr></thead>")
            f.write(sm_summary)
            f.write("</table>")
            if criterias_comment:
                f.write("<div class=\"note\">Note:\n"+criterias_comment+"</note>")
            f.write("</body></html>")
        # add the summary to the report at the first position
        supermetrics_report = [] # only keep the summary
        supermetrics_report.insert(0,AutoReportElement(title="supermetrics summary",path="supermetrics_summary.html",description=""))
    
    # write report
    # report abacustest.html
    if os.path.isfile(os.path.join(work_path, save_path, "abacustest.html")):
        reports.append(ReportSection(title="abacustest report",
                       elements=[AutoReportElement(title="abacustest report", path=os.path.join(save_path,"abacustest.html"), description="")], ncols=1,)) 
        
    # 2. supermetrics report
    #print("supermetrics_save:",supermetrics_save)
    if supermetrics_report:
        reports.append(ReportSection(title="supermetrics",
                       elements=supermetrics_report, ncols=1,
                       metrics=supermetrics_save,
                       tags=gen_sm_tag(allparams)))

    # 4. supermetrics special section
    if supermetrics_special_section:
        for i in supermetrics_special_section:
            reports.append(i)
            
    # produce the report
    # 1. metrics report
    if metrics_report:
        reports.append(ReportSection(title="metrics",
                       elements=metrics_report, ncols=1))
        
    # 3. metrics chart
    if metrics_chart_elements:
        reports.append(ReportSection(title="metrics chart",
                       elements=metrics_chart_elements, ncols=2))

    return reports


from ..model import Model
import os, glob, json, traceback
from . import comm
import numpy as np
from abacustest.outresult import pandas_out
from abacustest.report import gen_html
from abacustest.lib_report.table import output_float
from .comm_plot import auto_set_yaxis
from abacustest.lib_collectdata.collectdata import RESULT
import copy


class SPTestModel(Model):
    @staticmethod
    def model_name(): # type: ignore
        '''
        Name of the model, which will be used as the subcommand
        '''
        return "sptest"
    
    @staticmethod
    def description(): # type: ignore
        '''
        Description of the model
        '''
        return "Show the results of FP test"
    
    @staticmethod
    def add_args(parser):
        parser.add_argument("-j","--job",default=[], action="extend",nargs="*" ,help='the path of calculated jobs, or a json file containing results. If the jobpath is in json file, the results in job path is prefered.')
        parser.add_argument("--jobtype", default="abacus",type=str, help="the job type, should be abacus/qe/vasp. Default is abacus")
        parser.add_argument("--ref",type=str,default=None, help="the reference results in json file",)
        parser.add_argument("--testn",type=str,default="TEST", help="the label of test jobs",)
        parser.add_argument("--refn",type=str,default="REF", help="the label of reference results",)
        parser.add_argument("--remove_none_point", action="store_true", help="if remove the point where one of the value is None")
        parser.add_argument("--dpi", type=int, default=200, help="the dpi of the output image")

        
    def run(self,params):
        print(params)
        cm = SPTest(params.job, params.jobtype, params.ref, params.testn, params.refn,
                    remove_none_point=params.remove_none_point,
                    dpi=params.dpi)
        cm.run()
        
                    
class SPTest:
    def __init__(self, jobs, jobtype, ref, testn, refn,
                 remove_none_point=False,
                 dpi=200):
        self.test_r = self.read_test(jobs, jobtype)
        self.ref_r = self.read_reference(ref)
        self.testn = testn
        self.refn = refn
        self.remove_none_point = remove_none_point # if remove the point where one of the value is None
        self.dpi = dpi
    
    def run(self):
        if len(self.test_r) == 0:
            print("No results found")
            return
        # save test_r to metrics.json
        json.dump(self.test_r, open("metrics.json", "w"), indent=4)
        
        # 1. gen report.json
        report, cretria_set = self.gen_report()
        json.dump(report, open("report.json", "w"), indent=4)
        
        # 2. plot the statistical results
        self.plot_report(report, cretria_set, "acc.png", "perf.png", 
                         remove_none_point=self.remove_none_point,
                         dpi=self.dpi)
        json.dump({"acc_results":{"type": "image", "file": "acc.png"},
                   "per_results": {"type": "image", "file": "perf.png"}}, open("supermetrics.json", "w"), indent=4)
        
        # 3. generate the abacus.html
        gen_html({"content": [{"type": "metrics",
                               "content": "report.json",
                               "title": "Table of results.",
                               "criteria": cretria_set},
                              {"type": "image","content": "acc.png"},
                              {"type": "image","content": "perf.png"}]
                  },
                 "abacustest.html")
        
    
    def clean_example(self, r):
        return {k.rstrip("/"): r[k] for k in r.keys()}  
    
    def write_testr(self, fname="metrics.json"):
        json.dump(self.test_r, open(fname, "w"), indent=4)
    
    def read_reference(self, ref):
        if ref is None:
            return None
        else:
            try:
                return self.clean_example(json.load(open(ref)))
            except:
                traceback.print_exc()
                return None
    
    def read_test(self, jobs, jobtype):
        job_path = []
        result_file = []
        for job in jobs:
            if os.path.isfile(job) and job.endswith(".json"):
                result_file.append(job)
            elif os.path.isdir(job):
                job_path.append(job)
            else:
                continue
    
        r = {}
        for i in result_file:
            try:
                r.update(json.load(open(i)))  
            except:
                traceback.print_exc()
                print(f"Error in reading {i}, skip it")
            
        for i in job_path:
            try:
                print(i)
                result = RESULT(path = i, fmt = jobtype)
                ir = {key: result[key] for key in ["natom","label","nelec","normal_end","nkstot","ibzk", "converge", "volume","total_time", "scf_steps", 
                                                   "energy", "force", "stress", "virial", "ds_mag_force"]}
                r.update({i: ir})
            except:
                traceback.print_exc()
                print(f"Error in reading {i}, skip it")
        
        return self.clean_example(r)
    
    
    def gen_label(self, ilabel):
        if ilabel is None:
            return None
        elif isinstance(ilabel, str):
            return ilabel
        elif isinstance(ilabel, list):
            label_uniq = []
            for i in ilabel:
                if i not in label_uniq:
                    label_uniq.append(i)
            return "".join([i+str(ilabel.count(i)) for i in label_uniq])
        else:
            return None
    
    def cal_dev(self, r1, r2):
        if None in [r1, r2]:
            return None
        if isinstance(r1, (int, float)) and isinstance(r2, (int, float)):
            return abs(r1 - r2)
        if isinstance(r1, list) and isinstance(r2, list):
            r1 = np.array(r1).flatten()
            r2 = np.array(r2).flatten()
            if r1.shape != r2.shape:
                return None
            return np.max(np.abs(r1 - r2)).tolist()
            
    def remove_none(self, r):
        new_r = {}
        value_keys = list(r.values())[0].keys()
        values = {key: all([r[i][key] is None for i in r]) for key in value_keys}
        
        for example, value in r.items():
            new_r[example] = {key: value[key] for key in value_keys if not values[key]}
        return new_r
      
    def sort_report(self, r, key_list):
        new_r = {}
        for example, value in r.items():
            new_r[example] = {key: value[key] for key in key_list if key in value}
        return new_r              
        
    def gen_report(self):
        report = {}
        has_mf = False
        for example, result in self.test_r.items():
            report[example] = {
                "label": self.gen_label(result.get("label")),
                "natom": result.get("natom"),
                "nelec": result.get("nelec"),
                "nkpoint": result.get("nkstot"),
                "ibzk": result.get("ibzk"),
                "volume(A^3)": result.get("volume"),
                "NormalEnd": result.get("normal_end"),
                "Converge": result.get("converge"),
                "TotTime(s)": result.get("total_time") if result.get("normal_end") else None,
                "SCFSteps": result.get("scf_steps"),
            }
            if result.get("ds_mag_force") is not None:
                has_mf = True
            
            if self.ref_r is not None:
                for key,key_result in zip(["NormalEnd", "Converge", "TotTime(s)", "SCFSteps"],
                                          ["normal_end", "converge", "total_time", "scf_steps"]):
                    report[example][f"{key}({self.testn})"] = report[example].pop(key)
                    report[example][f"{key}({self.refn})"] = self.ref_r.get(example, {}).get(key_result)
                if report[example][f"NormalEnd({self.testn})"] in [None, False]:
                    report[example][f"TotTime(s)({self.testn})"] = None
                
                for key, name in zip(["energy", "force", "stress", "ds_mag_force"],
                                     ["DevE(eV)", "MaxDevF(eV/A)", "MaxDevS(kbar)", "MaxDevMagF(eV/uB)"]):
                    report[example][name] = self.cal_dev(result.get(key), self.ref_r.get(example, {}).get(key))
                
                if None in [report[example]["natom"], report[example]["DevE(eV)"]]:
                    report[example]["DevE(meV/atom)"] = None
                else:
                    report[example]["DevE(meV/atom)"] = report[example]["DevE(eV)"] * 1000 / report[example]["natom"]
            
            for k1, k2, k3 in [["TotTime(s)", "SCFSteps", "Time/step(s)"],
                                  [f"TotTime(s)({self.testn})", f"SCFSteps({self.testn})", f"Time/step(s)({self.testn})"],
                                  [f"TotTime(s)({self.refn})", f"SCFSteps({self.refn})", f"Time/step(s)({self.refn})"]]:
                if None not in [report[example].get(k1), report[example].get(k2)]:
                    report[example][k3] = report[example][k1]/report[example][k2]
                else:
                    report[example][k3] = None

        # remove the key with all None
        report = self.remove_none(report)
        # now sort the keys, if key is not in report, then it will be ignored
        sort_keys = ["label", "natom", "nelec", "volume(A^3)",
                     "NormalEnd", "Converge", "TotTime(s)","SCFSteps", "Time/step(s)",
                     f"NormalEnd({self.testn})", f"Converge({self.testn})", f"TotTime(s)({self.testn})", f"SCFSteps({self.testn})", f"Time/step(s)({self.testn})",
                     f"NormalEnd({self.refn})", f"Converge({self.refn})", f"TotTime(s)({self.refn})", f"SCFSteps({self.refn})", f"Time/step(s)({self.refn})",
                     "DevE(eV)", "DevE(meV/atom)", "MaxDevF(eV/A)", "MaxDevS(kbar)"]
        cretria_set = {
            "NormalEnd": "bool(x)",
            "Converge": "bool(x)",
            f"NormalEnd({self.testn})": "bool(x)",
            f"Converge({self.testn})": "bool(x)",
            "DevE(meV/atom)": "x < 1",
            "MaxDevF(eV/A)": "x < 0.01",
            "MaxDevS(kbar)": "x < 1",
        }
        if has_mf:
            sort_keys.append("MaxDevMagF(eV/uB)")
            cretria_set["MaxDevMagF(eV/uB)"] = "x < 0.01"
        report = self.sort_report(report, sort_keys)
        metric_keys = list(report[list(report.keys())[0]].keys())
        
        cretria_set = {k: v for k, v in cretria_set.items() if k in metric_keys}
        return report, cretria_set
    
    def cal_row_col(self, nfig, nexample):
        ncol = min(2, nfig)
        nrow = nfig//ncol
        while nrow*ncol < nfig: nrow += 1   
        width = max(6, 0.4*nexample)
        if width > 12:
            ncol = 1
            nrow = nfig
                
        return nrow, ncol
    
    def plot_report(self, report, cretria_set, acc_fname="acc.png", perf_fname="perf.png",remove_none_point=False, dpi=200):
        # remove_none_point: if remove the point where one of the value is None
        metric_keys = list(report[list(report.keys())[0]].keys())
        example_name = list(report.keys())
        example_idx = np.arange(len(example_name))
        
        # accuracy results
        acc_keys = ["DevE(meV/atom)", "MaxDevF(eV/A)", "MaxDevS(kbar)"]
        if "MaxDevMagF(eV/uB)" in metric_keys:
            acc_keys.append("MaxDevMagF(eV/uB)")
        acc_data = [None if i not in metric_keys else [report[k][i] for k in example_name] for i in acc_keys]
        acc_ifplot = [True if i is not None else False for i in acc_data]
        
        # performance results
        per_keys = ["TotTime(s)", "SCFSteps", "Time/step(s)"]
        if self.ref_r is not None:
            per_data1 = [None if f"{i}({self.testn})" not in metric_keys else [report[k][f"{i}({self.testn})"] for k in example_name] for i in per_keys]
        else:
            per_data1 = [None if i not in metric_keys else [report[k][i] for k in example_name] for i in per_keys]
        per_data2 = [None if f"{i}({self.refn})" not in metric_keys else [report[k][f"{i}({self.refn})"] for k in example_name] for i in per_keys]
        per_ifplot = [True if per_data1[i] is not None else False for i in range(len(per_keys))]
        
        nfig = sum(acc_ifplot) + sum(per_ifplot)
        if nfig == 0:
            print("No data to plot")
            return None, None
        
        import matplotlib.pyplot as plt
        fontsize = 18
        width = max(6, 0.4*len(example_name))
        
        # plot accuracy results
        nfig = sum(acc_ifplot)
        if nfig > 0:
            nrow, ncol = self.cal_row_col(nfig, len(example_name))
            fig, axs = plt.subplots(nrow,ncol,figsize=(width*ncol,6*nrow))
            axs = np.array(axs).flatten()
            for i in range(len(acc_keys)):
                if acc_ifplot[i]:
                    example_idx_tmp = copy.deepcopy(example_idx)
                    example_name_tmp = copy.deepcopy(example_name)
                    ax = axs[i]
                    criteria_line = float(cretria_set[acc_keys[i]].split("<")[-1])
                    x, y,e_tmp = comm.clean_none_list(example_idx, acc_data[i], example_name)
                    
                    if remove_none_point:
                        x = np.arange(len(y))
                        example_idx_tmp = x
                        example_name_tmp = e_tmp
                        
                    ax.bar(x, y, label=acc_keys[i])
                    ax.set_xticks(example_idx_tmp)
                    ax.set_xticklabels(example_name_tmp, rotation=30, fontsize=fontsize-2, ha="right")
                    ax.set_title(acc_keys[i], fontsize=fontsize)
                    #ymin, ymax = auto_set_yaxis(ax, y, log_scale_threshold=1000)
                    ax.axhline(criteria_line, color="red", linestyle="--", label=f"{criteria_line}")
                    ax.set_yscale("log")
                    
                    ax.set_xlabel("Example", fontsize=fontsize)
                    ax.set_ylabel(acc_keys[i], fontsize=fontsize)
                    ax.legend(fontsize=fontsize-2)
            plt.tight_layout()
            plt.savefig(acc_fname,dpi=dpi)
            plt.close()
        
        nfig = sum(per_ifplot)
        if nfig > 0:
            nrow, ncol = self.cal_row_col(nfig, len(example_name))
            fig, axs = plt.subplots(nrow,ncol,figsize=(width*ncol,6*nrow))
            axs = np.array(axs).flatten()
            for i in range(len(per_keys)):
                if per_ifplot[i]:
                    ax = axs[i]
                    example_idx_tmp = copy.deepcopy(example_idx)
                    example_name_tmp = copy.deepcopy(example_name)
                    if per_data2[i] is None:
                        x, tot_y,e_tmp = comm.clean_none_list(example_idx, per_data1[i], example_name)
                        if remove_none_point:
                            x = np.arange(len(tot_y))
                            example_idx_tmp = x
                            example_name_tmp = e_tmp
                        ax.bar(x, tot_y, label=f"{self.testn}")
                    else:
                        if remove_none_point:
                            x1, y1, x2, y2, e_tmp = comm.clean_none_list(example_idx, per_data1[i], example_idx, per_data2[i], example_name)
                            x1 = x2 = np.arange(len(y1))
                            example_idx_tmp = x1
                            example_name_tmp = e_tmp
                        else:
                            x1, y1 = comm.clean_none_list(example_idx, per_data1[i])
                            x2, y2 = comm.clean_none_list(example_idx, per_data2[i])
                        
                        ax.bar(np.array(x1)-0.2, y1, label=f"{self.testn}", width=0.4)
                        ax.bar(np.array(x2)+0.2, y2, label=f"{self.refn}", width=0.4)
                        tot_y = y1 + y2

                    ax.set_xticks(example_idx_tmp)
                    ax.set_xticklabels(example_name_tmp, rotation=30, fontsize=fontsize-2, ha="right")
                    ax.set_title(per_keys[i], fontsize=fontsize)
                    ax.set_xlabel("Example", fontsize=fontsize)
                    ax.set_ylabel(per_keys[i], fontsize=fontsize)
                    ax.legend(fontsize=fontsize-2, loc="upper left")
                    auto_set_yaxis(ax, tot_y,log_scale_threshold=1000)

                    if per_data2[i] is not None:
                        ratio = [None if per_data1[i][j] is None or per_data2[i][j] is None else per_data1[i][j]/per_data2[i][j] for j in range(len(example_name_tmp))]
                        x, y = comm.clean_none_list(example_idx_tmp, ratio)
                        if len(y) == 0: continue
                        # calculate the geometric mean of ratio
                        ratio = np.prod(y)**(1/len(y))
                        axs1 = ax.twinx()
                        axs1.plot(x, y, color="red", linestyle="--", label=f"{self.testn}/{self.refn}={output_float(ratio)}")
                        axs1.legend(fontsize=fontsize-2, loc="upper right")
                        axs1.set_ylabel("Ratio", fontsize=fontsize)
                        auto_set_yaxis(axs1, y,log_scale_threshold=1000)
            plt.tight_layout()
            plt.savefig(perf_fname,dpi=dpi)
            plt.close()
        return acc_fname, perf_fname
        
        
        
        
    
        
        
                    
                
                
        
        
        

    


import os,sys,glob,re
from ..resultVasp import ResultVasp
from .. import comm
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element

xfmlt = comm.XmlFindMultiLayerText
xfml = comm.XmlFindMultiLayer

class Vasp(ResultVasp):
    
    @ResultVasp.register(version="the vasp version",
                         ncore = "mpi cores")
    def GetGeneralInfo(self):
        if self.XMLROOT != None:
            self['version'] = self.XMLROOT.find("./generator/i[@name='version']").text
        
        for line in self.OUTCAR:
            if "running on" in line and "total cores" in line:
                self['ncore'] = int(line.split()[2])
    
    @ResultVasp.register(normal_end="if the job is normal ending")
    def GetNormalEnd(self):
        if len(self.OUTCAR) == 0:
            self['normal_end'] = None
            return
        elif len(self.OUTCAR) > 0 and "Voluntary context switches:" in self.OUTCAR[-1]:
            self['normal_end'] = True
            return
        else:
            self['normal_end'] = False

            print("Job is not normal ending!!! The latest 10 lines is:")
            if len(self.OUTCAR) < 10:
                print(''.join(self.OUTCAR))
            else:
                print(''.join(self.OUTCAR[-10:]))

    @ResultVasp.register(kpt="list, the K POINTS setting",
                         nkstot = "total K point number",
                         ibzk = "irreducible K point number")
    def GetKPT(self):
        if self.XMLROOT != None:
            tree = "./kpoints/generation/v[@name='divisions']"
            kpts = self.XMLROOT.find(tree)
            if kpts != None:
                nkstot = 1
                kpt = []
                for i in kpts.text.split():
                    nkstot *= int(i)
                    kpt.append(int(i))
                self['kpt'] = kpt
                self['nkstot'] = nkstot
            else:
                self['kpt'] = None
                self['nkstot'] = None

            tree = "./kpoints/varray[@name='kpointlist']/v"
            ibzk = self.XMLROOT.findall(tree)
            if ibzk != None:
                self['ibzk'] = len(ibzk)
            else:
                self['ibzk'] = None
        else:
            ibzk = None
            kpt = None
            nkstot = None
            for i,line in enumerate(self.OUTCAR):
                if "k-points in BZ" in line:
                    ibzk = int(line.split()[3])
                elif " generate k-points for:" in line:
                    kpt = [int(j) for j in self.OUTCAR[i].split()[-3:]]
                elif "k-points in reciprocal lattice and weights" in line:
                    nkstot = 0
                    for j in range(i+1,len(self.OUTCAR)):
                        if self.OUTCAR[j].strip() == "":
                            break
                        nkstot += 1
                    break
                    
            self["ibzk"] = ibzk        
            self['kpt'] = kpt
            self['nkstot'] = nkstot        
                
    @ResultVasp.register(nbands="number of bands",
                         nelec = "total electron number",
                         spin = "the spin number",
                         encut = "eV, the energy cutoff",
                         ismear = "the smearing method",
                         sigma = "the SIGMA setting of smearing, in eV",
                         nelm = "value of NELM, the setted maximum SCF steps",
                         natom = "total atom number",
                         volume = "volume(A^3). if is relax or md, will return the volume of last ION step")
    def GetInputSetting(self):
        if self.XMLROOT != None:
            tree = ".//separator[@name='electronic']/i[@name='NBANDS']"
            self['nbands'] = comm.iint(comm.XmlGetText(self.XMLROOT.find(tree)))

            tree = ".//separator[@name='electronic']/i[@name='NELECT']"
            self['nelec'] = comm.ifloat(comm.XmlGetText(self.XMLROOT.find(tree)))

            tree = ".//separator[@name='electronic spin']/i[@name='ISPIN']"
            self['spin'] = comm.iint(comm.XmlGetText(self.XMLROOT.find(tree)))
            
            tree = ".//separator[@name='electronic']/i[@name='ENMAX']"
            self['encut'] = comm.ifloat(comm.XmlGetText(self.XMLROOT.find(tree)))

            tree = ".//separator[@name='electronic smearing']/i[@name='ISMEAR']"
            self['ismear'] = comm.iint(comm.XmlGetText(self.XMLROOT.find(tree)))

            tree = ".//separator[@name='electronic smearing']/i[@name='SIGMA']"
            self['sigma'] = comm.ifloat(comm.XmlGetText(self.XMLROOT.find(tree)))
            
            tree = ".//separator[@name='electronic convergence']/i[@name='NELM']"
            self['nelm'] = comm.iint(comm.XmlGetText(self.XMLROOT.find(tree)))

            tree = "./atominfo/atoms"
            self['natom'] = comm.iint(comm.XmlGetText(self.XMLROOT.find(tree)))
            
            tree = "./structure/crystal/i[@name='volume']"
            self['volume'] = comm.ifloat(comm.XmlGetText(self.XMLROOT.findall(tree),idx=-1))
            
        else:
            volume = None
            for line in self.OUTCAR:
                sline = line.split()
                if "number of bands    NBANDS" in line:
                    self['nbands'] = int(sline[-1])
                elif "number of ions     NIONS =" in line:
                    self['natom'] = int(sline[-1])
                elif "ISPIN  =" in line:
                    self['spin'] = int(sline[2])
                elif "ENCUT  =  " in line:
                    self['encut'] = float(sline[4])
                elif "NELECT =" in line:
                    self["nelec"] = float(sline[2])
                elif "ISMEAR =" in line:
                    self['ismear'] = sline[2].rstrip(";")
                    self["sigma"] = float(sline[5])
                elif "NELM   =" in line:
                    self['nelm'] = int(sline[2][:-1])
                elif "volume of cell" in line:
                    volume = float(sline[-1])
            self["volume"] = volume

    @ResultVasp.register(ldautype = "value of LDAUTYPE, the type of plus U",
                         ldaul = "list, value of LDAUL, the l-quantum number of each element",
                         ldauu = "list, value of LDAUU, the U setting",
                         ldauj = "list, value of LDAUJ, the J setting")
    def GetLdaUSetting(self):
        for i,line in enumerate(self.OUTCAR):
            sline = line.split()
            if "LDA+U is selected," in line:
                self['ldautype'] = int(sline[-1])
                self['ldaul'] = [int(j) for j in self.OUTCAR[i+1].split("=")[-1].split()]
                self['ldauu'] = [float(j) for j in self.OUTCAR[i+2].split("=")[-1].split()]
                self['ldauj'] = [float(j) for j in self.OUTCAR[i+3].split("=")[-1].split()]
                break

    @ResultVasp.register(denergy_last = 'The energy difference in SCF, eV',
                         denergy = "The last energy difference in SCF, eV")
    def GetDE(self):
        des = []
        for i in range(len(self.OSZICAR)):
            line = self.OSZICAR[i]
            if line.strip() == "": continue
            if line[3] == ":" and len(line.split()) == 7:
                de = float(line.split()[3])
                des.append(de)
        if len(des) > 0:
            self['denergy_last'] = des[-1]
            self['denergy'] = des
        else:
            self['denergy_last'] = None
            self['denergy'] = None
   
    @ResultVasp.register(scf_steps = 'the steps of SCF, if is relax or md job, only last ION step is read',
                         converge = "if the SCF is converged. If scf_steps is smaller than NELM, will be converged, else is not converged")
    def GetSCFInfo(self):
        for i in range(len(self.OUTCAR)):
            j = -1*i - 1
            line = self.OUTCAR[j]
            if 'Iteration' in line:
                self['scf_steps'] = int(line.split('(')[1].split(')')[0])
                if self['scf_steps'] < self['nelm'] and self['efermi'] is not None:
                    # in some case the SCF is not converged but terminated, so we need to check the fermi energy
                    self['converge'] = True
                else:
                    self['converge'] = False
                break
    
    @ResultVasp.register(energy = 'eV,the total energy, if is relax or md job, will return the energy of last ION step',
                         energy_per_atom = 'eV, the energy divided by natom, if is relax or md job, will return the energy of last ION step')
    def GetEnergy(self):
        for i in range(len(self.OUTCAR)):
            line = self.OUTCAR[-i-1]
            if "energy  without entropy=" in line:
                self['energy'] = float(line.split()[-4])
                if self['natom'] != None:
                    self['energy_per_atom'] = self['energy'] / self['natom']
                else:
                    self['energy_per_atom'] = None
                return
    
    @ResultVasp.register(force = 'list, eV/angstrom, the force of all atoms, [atom1x,atom1y,atom1z,atom2x,atom2y,atom2z...]',
                         stress = 'list, kBar, the stress, [xx,xy,xz,yx,yy,yz,zx,zy,zz]',
                         virial='list, eV, the virial, [xx,xy,xz,yx,yy,yz,zx,zy,zz]',
                         pressure="kBar, the pressure, 1/3*trace(stress)",
                         cell = 'list[list], Angstrom, the vector of cell. If is RELAX or MD job, will output the last cell.',
                         )
    def GetForceStress(self):
        force = None
        stress = None
        virial = None
        cell = None
        for i,line in enumerate(self.OUTCAR):
            if 'TOTAL-FORCE (eV/Angst)' in line:
                j = i+2
                force = []
                while self.OUTCAR[j][:3] != " --":
                    force += [float(k) for k in self.OUTCAR[j].split()[3:6]]
                    j += 1
            elif '  in kB' in line:
                s = [float(i) for i in line.split()[2:8]]
                stress = [s[0],s[3],s[5],s[3],s[1],s[4],s[5],s[4],s[2]]
                v = [float(i) for i in self.OUTCAR[i-1].split()[1:7]]
                virial = [v[0],v[3],v[5],v[3],v[1],v[4],v[5],v[4],v[2]]
            elif "VOLUME and BASIS-vectors are now :" in line:
                cell = []
                for j in range(3):
                    cell.append([float(k) for k in self.OUTCAR[i+j+5].split()[0:3]])
        self['force'] = force
        self['stress']  = stress
        self['virial'] = virial
        self["cell"] = cell
        if stress:
            self['pressure'] = 1.0/3.0*(stress[0]+stress[4]+stress[8])
        else:
            self['pressure'] = None
        
    @ResultVasp.register(total_time = 'Total CPU time (s)',
                         scf_time = 'the total SCF times, s',
                         stress_time = 'the time of calculating stress')                         
    def GetTimeInfo(self):
        stresst = None
        scft = 0
        for line in self.OUTCAR:
            if 'STRESS:  cpu time' in line:
                stresst = float(line.split()[-1])
            elif 'LOOP:  cpu time' in line:
                scft += float(line.split()[-1])
            elif 'Total CPU time used (sec):' in line:
                self['total_time'] = float(line.split()[-1])

        if stresst != None:
            self['stress_time'] = stresst
        if scft > 0:
            self['scf_time'] = scft

    @ResultVasp.register(total_mag = 'total magnization',
                         absolute_mag="absolute magnetism, the summation of the magnetic moment of all atoms",
                         atom_mag = 'list, the magnization of each atom')
    def GetMagInfo(self):
        getatommag = False
        atommagsx = []
        atommagsy = []
        atommagsz = []
        
        for i in range(len(self.OUTCAR)):
            i = -i-1
            line = self.OUTCAR[i]
            if line[:19] == " number of electron":
                self['total_mag'] = float(line.split()[-1])
                break
            elif line[:18] == ' magnetization (x)':
                j = i + 4
                atommag = []
                while self.OUTCAR[j][:3] != "---":
                    atommag.append(float(self.OUTCAR[j].split()[-1]))
                    j += 1
                atommagsx.append(atommag)
            elif line[:18] == ' magnetization (y)':
                j = i + 4
                atommag = []
                while self.OUTCAR[j][:3] != "---":
                    atommag.append(float(self.OUTCAR[j].split()[-1]))
                    j += 1
                atommagsy.append(atommag)
            elif line[:18] == ' magnetization (z)':
                j = i + 4
                atommag = []
                while self.OUTCAR[j][:3] != "---":
                    atommag.append(float(self.OUTCAR[j].split()[-1]))
                    j += 1
                atommagsz.append(atommag)
        
        if len(atommagsx) > 0 and len(atommagsy) > 0 and len(atommagsz) > 0:
            assert len(atommagsx) == len(atommagsy) == len(atommagsz), "The atom magnetism in x, y, z direction should have the same length."
            atomx = atommagsx[0]
            atomy = atommagsy[0]
            atomz = atommagsz[0]
            atommag = []
            abs_mag = 0
            for i in range(len(atomx)):
                atommag.append([atomx[i],atomy[i],atomz[i]])
                abs_mag += (atomx[i]**2 + atomy[i]**2 + atomz[i]**2)**0.5
            self["atom_mag"] = atommag
            self["absolute_mag"] = abs_mag
        elif len(atommagsx) > 0:
            self["atom_mag"] = atommagsx[0]
            self["absolute_mag"] = sum([abs(i) for i in atommagsx[0]])
        else:
            self["atom_mag"] = None
            self["absolute_mag"]
                
    
    @ResultVasp.register(atom_name = 'list, the element name of each atom' ,
                         atom_type = 'list, the element name of each atomtype',
                         element_list = 'list, the element name of all atoms',
                         element = "list[], a list of the element name of all atoms",
                           label = "list[], a list of atom label of all atoms",
                           atomlabel_list = "same as label",
                         efermi     = 'the fermi energy, eV')
    def GetXMLInfo(self):
        if self.XMLROOT != None:
            atom_name = comm.XmlGetText(self.XMLROOT.findall("./atominfo/array[@name='atoms']/set/rc/c[1]"))
            if isinstance(atom_name,list):
                atom_name = [i.strip() for i in atom_name]
            self["atom_name"] = atom_name
            
            atom_type = comm.XmlGetText(self.XMLROOT.findall("./atominfo/array[@name='atomtypes']/set/rc/c[2]"))
            if isinstance(atom_type,list):
                atom_type = [i.strip() for i in atom_type]
            self["atom_type"] = atom_type
            
            self['efermi'] = comm.XmlGetText(self.XMLROOT.findall("./calculation/dos/i[@name='efermi'][last()]"),func=float,idx = -1)
            element = []
            for i in self.XMLROOT.findall("./atominfo/array[@name='atoms']/set/rc/c[1]"):
                element.append(i.text)
            self["element"] = element
            self["label"] = element
            self["element_list"] = element
            self["atomlabel_list"] = element
        else:
            # read from OUTCAR
            atom_name = None
            atom_type = None
            efermi = None
            for i in range(len(self.OUTCAR)):
                if "VRHFIN" in self.OUTCAR[i]:
                    if atom_name == None:
                        atom_name = []
                    atom_name.append(self.OUTCAR[i].split(":")[0].split("=")[1].strip())
                elif "E-fermi" in self.OUTCAR[i]:
                    efermi = float(self.OUTCAR[i].split()[2])
            if atom_name != None:
                atom_type = []
                for i in atom_name:
                    if i not in atom_type:
                        atom_type.append(i)
            self['atom_type'] = atom_type
            self['atom_name'] = atom_name
            self["element"] = atom_name
            self["label"] = atom_name
            self["element_list"] = atom_name
            self["atomlabel_list"] = atom_name
            self['efermi'] = efermi

    @ResultVasp.register(band = '[[[]]], list with three dimension spin*kpoint*band')
    def GetBandInfo(self):
        if self.XMLROOT != None:
            band = []
            eigen = self.XMLROOT.findall('./calculation/eigenvalues/array')
            if eigen == None:
                self['band'] = None
            else:
                array = eigen[-1]
                for spin in array.find('set').findall('set'):
                    band.append([])
                    for kpoint in spin.findall('set'):
                        band[-1].append([])
                        for iband in kpoint.findall('r'):
                            band[-1][-1].append(float(iband.text.split()[0]))
                self['band'] = band
        else:
            # try to read band from OUTCAR
            band = []
            nspin = self["spin"]
            nk = self["nkstot"]
            nband = self["nbands"]
            if nspin == None or nk == None:
                self['band'] = None
                return
            print("nspin=%d, nk=%d, nband=%d" % (nspin,nk,nband))
            for i in range(len(self.OUTCAR)):
                if "spin component" in self.OUTCAR[i]:
                    band.append([])
                    for j in range(nk):
                        nk_start = i + 1 + (nband+3)*j + 3
                        nk_end = nk_start + nband
                        band[-1].append([float(k.split()[1]) for k in self.OUTCAR[nk_start:nk_end]])
                        
                    nspin -= 1
                    if nspin == 0:
                        break
            if len(band) == 0:
                self['band'] = None
            else:
                self['band'] = band
    
    @ResultVasp.register(band_gap = 'eV, the band gap')
    def GetBandGap(self):
        if self['band'] == None or self['efermi'] == None:
            self['band_gap'] = None
        else:
            vb = None
            cb = None
            fermi = self['efermi']
            self["band_gap"] = comm.cal_band_gap(self['band'],fermi)
    '''
    @ResultVasp.register(band_gap = 'eV, the band gap')
    def GetBandGap(self):
        if self['band'] == None or self['nelec'] == None:
            self['band_gap'] = None
        else:
            occu_band = int(self['nelec'] / 2)
            band_num = len(self['band'][0][0])
            if occu_band <= 0:
                self['band_gap'] = None
                return
            if occu_band >= band_num:
                self['band_gap'] = None
                return
             
            vb = None
            cb = None
            for ispin in self['band']:
                for ik in ispin:
                    eband1 = ik[occu_band-1]
                    eband2 = ik[occu_band]
                    if cb == None or cb < eband1:
                        cb = eband1
                    if vb == None or vb > eband2:
                        vb = eband2 
            #print("cb=%.5f, vb=%.5f" % (cb,vb))
            if vb == None or cb == None:
                self['band_gap'] = None
                return
            else:
                band_gap = vb - cb
                if band_gap < 0:
                    band_gap = 0
                self['band_gap'] = band_gap
                return
    '''
    @ResultVasp.register(point_group = 'point group',
                         point_group_in_space_group = "point group in space group")
    def GetPointGroup(self):
        pg = None
        pgsg = None
        for i in self.OUTCAR:
            if pg != None and pgsg != None:
                break
            
            if "The dynamic configuration has the point symmetry" in i:
                pg = i.strip()[48:-1].strip()
            elif " The point group associated with its full space group is" in i:
                pgsg = i.strip()[56:-1].strip()
        self['point_group'] = pg
        self['point_group_in_space_group'] = pgsg
        
    @ResultVasp.register(relax_steps = 'the steps of ION',
                         relax_converge = 'if the relax is converged',)
    def GetRelaxInfo(self):
        if self.XMLROOT != None:
            self["relax_steps"] = len(self.XMLROOT.findall("calculation"))
            
            #max_ion = comm.iint(self.XMLROOT.findtext("./parameters/ionic[@name='NSW']",default=64))
            max_ion = comm.iint(self.XMLROOT.findtext("./parameters/separator[@name='ionic']/i[@name='NSW']",default=64))
            if max_ion != None and self["relax_steps"] != None:
                if max_ion > self["relax_steps"]:
                    self["relax_converge"] = True
                else:
                    self["relax_converge"] = False
            else:
                self["relax_converge"] = None
        elif self.OUTCAR:
            nsw = 1
            for i in self.OUTCAR:
                if "   NSW    =" in i:
                    nsw = int(i.split()[2])
            relax_steps = 0
            for i in self.OUTCAR[::-1]:
                if "--------------------------------------- Iteration" in i:
                    relax_steps = int(i[49:56])
                    break
            if relax_steps == 0 or relax_steps == nsw:
                self["relax_converge"] = False
            else:
                self["relax_converge"] = True
            self["relax_steps"] = relax_steps


import os,json,glob,shutil,traceback
import subprocess,copy
from abacustest.lib_prepare.abacus import AbacusStru,ReadInput,ReadKpt
import select

BOHRIUM_DES = '''If you use Bohrium to accelerate the calculation, you need to set below environment variables:
    export BOHRIUM_USERNAME=<your username> BOHRIUM_PASSWORD=<your password> BOHRIUM_PROJECT_ID=<your project id> 
Or you can add below block in your setting.json:
    "config": {
        "bohrium_username": "<your username>",
        "bohrium_password": "<your password>",
        "bohrium_project_id": "<your project id>"
    } 
Or you can use dispatcher to submit the calculation to the remote server,
and please refer to the dispatcher section for more details in https://github.com/pxlxingliang/abacus-test/tree/develop.'''

def doc_after_prepare(model_name, jobs, files,has_prepare=True):
    print(f"\nPrepare the {model_name} successfully!!!")
    print("\nYou can choose one of below methods to do the calculation:")
    print("1. run with abacustest locally")
    print(BOHRIUM_DES,)
    print("After finishing the setting, you can run below command to submit the calculation:")
    print("        abacustest submit -p setting.json &")
    print("\n2. run with bohrium app abacustest:")
    print("Compress the inputs and submit to https://app.bohrium.dp.tech/abacustest/?request=GET%3A%2Fapplications%2Fabacustest with Reuse Submodel.")
    print("Compress command:")
    print("    tar -zcvf %s.tar.gz %s" % (model_name, " ".join(jobs+files)))
    if has_prepare:
        print("\n3. run with your own script:")
        print("You can execute below command and abacustest will generate all inputs in folder abacustest/, and then you can submit the calculation by yourself.")
        outfolder = "" if len(jobs) > 1 else f"-s {jobs[0]}"
        print("    abacustest prepare -p setting.json " + outfolder + "\n\n")
    
def dump_setting(setting):
    """
    Dump the given setting dictionary to a JSON file named 'setting.json'.
    If 'setting.json' already exists, it will be backed up with a numbered suffix.

    Args:
        setting (dict): The setting dictionary to be dumped.

    Returns:
        None
    """
    if os.path.isfile("setting.json"):
        i = 1
        settingbakf = f"setting.json.bak{i}"
        while os.path.isfile(settingbakf):
            i += 1
            settingbakf = f"setting.json.bak{i}"
        print("Warning: setting.json exists. Will backup it to", settingbakf)
        os.system(f"mv setting.json {settingbakf}")
    json.dump(setting, open("setting.json", "w"), indent=4)

def get_physical_cores():
    """
    Get the number of physical cores in the system.

    Returns:
        int: The number of physical cores.

    Raises:
        CalledProcessError: If the command execution fails.
    """
    cmd = "lscpu | grep 'Core(s) per socket:' | awk '{print $4}'"
    cores = subprocess.check_output(cmd, shell=True).decode().strip()
    return int(cores)

def get_job_list(jobs):
    '''
    Get a list of job directories from the given list of job patterns.

    Args:
        jobs (list): A list of job patterns.

    Returns:
        list: A list of job directories.
    '''
    job_list = []
    for job in jobs:
        for ijob in glob.glob(job):
            if os.path.isdir(ijob):                        
                job_list.append(ijob)
    return job_list
    
def gen_supermetrics(file_name, sm_name=None, image=True):
    # file_type should be image or html
    if sm_name == None:
        basename = os.path.basename(file_name.rstrip("/"))
        if "." in basename:
            sm_name = os.path.splitext(basename)[0]
        else:
            sm_name = basename
    return {
        sm_name: {"type": {True:"image", False: "html"}.get(bool(image)),
                  "file": file_name}
    } 

def bak_file(filename, bak_org=False):
    '''
    Check if the file exists, and if so:
    if bak_org is True, backup the original file to a non-exist filename.bak{n}, n = 1,2,3,...
    else, return the filename.bak{n} that does not exist.
    '''
    if not os.path.exists(filename):
        return filename
    
    i = 1
    bakf = f"{filename}.bak{i}"
    while os.path.exists(bakf):
        i += 1
        bakf = f"{filename}.bak{i}"
        
    if bak_org:
        os.system(f"mv {filename} {bakf}")
        return filename 
    else:
        return bakf

def clean_files(ipath, f_list=[], folder_list=[]):
    '''
    Clean the files in the given folder.
    '''
    for ifile in f_list:
        for jfile in glob.glob(os.path.join(ipath,ifile)):
            if os.path.isfile(jfile):
                os.remove(jfile)
    for ifolder in folder_list:
        for jfolder in glob.glob(os.path.join(ipath,ifolder)):
            if os.path.isdir(jfolder):
                shutil.rmtree(jfolder)       

def get_abacus_inputfiles(ipath):
    """
    Find the input files in the given path and return a dict:
    {
        "input": input_param, # a dict of the input parameters, kpt_file, stru_file, pseudo_dir, orb_dir will be popped out or modified
        "stru": stru, # an AbacusStru object or None. The pp/orb now is only the base name.
        "kptf": kptf, # the absolute path of the kpt file or None if not found
        "pp": pp_abs_path, # a list of the absolute path of the pseudopotential files or [] if defined in STRU
        "orb": orb_abs_path, # a list of the absolute path of the orbital files or [] if defined in STRU
        "extra_files": extra_files # a list of the absolute path of the extra files
    }
    
    1. the kptf/pp/orb files are the absolute path, but not check if the files exist
    2. if the INPUT or STRU is not read, the corresponding value will be None
    3. if INPUT defines the stru_file, pseudo_dir or orb_dir, the returned dict will delete the corresponding key in the input_param
    4. extra_files are the files in the folder but not in the list of INPUT, STRU, KPT, PP, ORB
    """
    if not os.path.isdir(ipath):
        print(f"ERROR: {ipath} is not a directory")
        return {
            "input": None,
            "stru": None,
            "kpt": None,
            "pp": [],
            "orb": [],
            "extra_files": []
        }
    
    pwd = os.getcwd()
    os.chdir(ipath)
    
    if os.path.isfile("INPUT"):
        input_param = ReadInput("INPUT")
        struf = input_param.pop("stru_file","STRU")
        kptf = "KPT"
        if "kpt_file" in input_param:
            kptf = input_param["kpt_file"]
            input_param["kpt_file"] = os.path.basename(kptf)
    else:
        print(f"ERROR: INPUT file not found in {ipath}")
        input_param = {}
        kptf = "KPT"
        struf = "STRU"
        
    if not os.path.isfile(struf):
        print(f"ERROR: STRU file '{struf}' not found in {ipath}")
        stru = None
    else:
        stru = AbacusStru.ReadStru(struf)
        if not stru:
            print(f"ERROR: read STRU failed in {ipath}")
            stru = None
            
    pp_path = input_param.pop("pseudo_dir","")
    orb_path = input_param.pop("orb_dir","")
    
    pp_name = []
    orb_name = []
    pp_abs_path = None
    orb_abs_path = None
    if stru:
        pp = stru.get_pp()
        orb = stru.get_orb()
        pp = pp if pp else []
        orb = orb if orb else []
        pp_abs_path = [os.path.abspath(os.path.join(pp_path,i)) for i in pp]
        orb_abs_path = [os.path.abspath(os.path.join(orb_path,i)) for i in orb]
        pp_name = [os.path.basename(i) for i in pp]
        orb_name = [os.path.basename(i) for i in orb]
        if len(pp_name) > 0:
            stru.set_pp(pp_name)
        if len(orb_name) > 0:
            stru.set_orb(orb_name)    
    
    extra_files = []
    for ifile in os.listdir("."):
        if ifile not in ["INPUT",struf,kptf]+pp_name+orb_name and os.path.isfile(os.path.join(ifile)):
            extra_files.append(os.path.abspath(ifile))
    
    kptf = os.path.abspath(kptf) if os.path.isfile(kptf) else None
    os.chdir(pwd)        
    return {
        "input": input_param if input_param else None,
        "stru": stru,
        "kpt": kptf,
        "pp": pp_abs_path,
        "orb": orb_abs_path,
        "extra_files": extra_files
    }

def clean_none_list(*args):
    '''
    For given same length lists, remove the elements that are None in the same position.
    If the length of the lists are different, return the original lists.
    '''
    if len(args) == 0:
        return []
    n = len(args[0])
    for i in args:
        if len(i) != n:
            return args
    new_args = [[] for i in range(len(args))]
    for i in range(n):
        if all([not isinstance(j[i],type(None)) for j in args]):
            for j in range(len(args)):
                new_args[j].append(args[j][i])
    return tuple(new_args)

def run_command(
        cmd,
        shell=True
):
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=shell,
        executable='/bin/bash'
    )
    out = ""
    err = ""
    while True:
        readable, _, _ = select.select(
            [process.stdout, process.stderr], [], [])

        # 读取已经准备好的输出
        for fd in readable:
            if fd == process.stdout:
                line = process.stdout.readline()
                print(line.decode()[:-1])
                out += line.decode()
            elif fd == process.stderr:
                line = process.stderr.readline()
                print("STDERR:", line.decode()[:-1])
                err += line.decode()

        # 如果子进程已经结束，则退出循环
        return_code = process.poll()
        if return_code is not None:
            break
    return return_code, out, err


from ..model import Model
from . import comm
import argparse,json, os
from abacustest.lib_prepare.abacus import WriteKpt, WriteInput, gen_stru, ReadInput, AbacusStru
from pathlib import Path
from abacustest.lib_model.model_012_band import PrepBand
from abacustest.constant import RECOMMAND_IMAGE

from typing import List, Dict, Any
from typing import Union
from pathlib import Path
from abacustest.constant import EV2RY
from abacustest.lib_prepare.comm import collect_pp
import warnings



class Vasp2AbacusModel(Model):
    @staticmethod
    def model_name(): # type: ignore
        '''
        Name of the model, which will be used as the subcommand
        '''
        return "vasp2abacus"
    
    @staticmethod
    def description(): # type: ignore
        '''
        Description of the model
        '''
        return "Transform VASP input files to ABACUS input files."
    
    @staticmethod
    def add_args(parser):
        '''
        Add arguments for the prepare subcommand
        The arguments can not be command, model, modelcommand '''
        
        parser.description = "Transform VASP input files to ABACUS input files. ENCUT and EDIFF will not be transformed."
        parser.add_argument('-j', '--job',default=[], action="extend",nargs="*" ,help='the paths of VASP jobs, should contain INCAR, POSCAR, or KPOINTS')
        parser.add_argument("--pp",default=None,type=str,help="the path of pseudopotential library, or read from enviroment variable ABACUS_PP_PATH")
        parser.add_argument("--orb",default=None,type=str,help="the path of orbital library, or read from enviroment variable ABACUS_ORB_PATH")
        parser.add_argument("--input", default=None, type=str, help="A template input file, used to set the default parameters in ABACUS INPUT. The setting in this file will overwrite the parameters transferred from VASP INCAR.")
        return parser
    
    def run(self,params):
        '''
        Run the model with the given parameters
        '''
        # read the parameters
        jobs = params.job
        pp_path = params.pp
        orb_path = params.orb
        
        if pp_path is None:
            pp_path = os.environ.get("ABACUS_PP_PATH", None)
        if orb_path is None:
            orb_path = os.environ.get("ABACUS_ORB_PATH", None)

        
        vasp2abacus = Vasp2Abacus(jobs, pp_path, orb_path, params.input)
        vasp2abacus.run()


class Vasp2Abacus():
    """Transform VASP input files to ABACUS input files.
    Parameters
    ----------
    jobs : List[Union[str,Path]]
        The paths of VASP jobs, will transfer INCAR to ABACUS INPUT, POSCAR to ABACUS STRU, and KPOINTS to ABACUS KPT.
        If file does not exist, then related transfer will be skipped.
    pp_path : Union[str,Path], optional
        The path of pseudopotential library, used to set the pseudopotentials in ABACUS STRU.
        If not set, no pseudopotentials will be set.
    orb_path : Union[str,Path], optional
        The path of orbital library, used to set the orbitals in ABACUS STRU.
        If not set, no orbitals will be set.
    input_file: Union[str,Path], optional
        The path of the template input file, used to set the default parameters in ABACUS INPUT.
        The setting in this file will overwrite the parameters transferred from VASP INCAR.
        If not set, the below default parameters will be used:
        - basis_type: "pw" if lcao is False, "lcao" if lcao is True
        - ecutwfc: 80 Ry for pw, 100 Ry for lcao.
    """
    def __init__(self, 
                 jobs: List[Union[str,Path]], 
                 pp_path: Union[str,Path]=None, 
                 orb_path: Union[str,Path]=None,
                 input_file: Union[str,Path]=None,
                 ):
        self.jobs = jobs
        self.pp_path = pp_path 
        self.orb_path = orb_path
        
        # read pp and orb path
        self.pp_lib = None
        self.orb_lib = None
        self.recommand_ecutwfc = None
        if pp_path is not None:
            self.pp_lib = collect_pp(pp_path)
            if os.path.isfile(os.path.join(pp_path, "ecutwfc.json")):
                self.recommand_ecutwfc = json.load(open(os.path.join(pp_path, "ecutwfc.json"), "r"))
        if orb_path is not None:
            self.orb_lib = collect_pp(orb_path)
        
        # read input file
        if input_file is not None:
            self.input_template = ReadInput(input_file)
        else:
            self.input_template = {} 
            
    def run(self):
        transfer_jobs = []
        for job in self.jobs:
            if not os.path.exists(job):
                raise ValueError(f"Job path {job} does not exist.")
            print(f"\nProcessing job {job}...")
            
            # transfer INCAR
            if os.path.isfile(os.path.join(job, "INCAR")):
                input_param, magmom = self.transfer_incar(os.path.join(job, "INCAR"))
            else:
                input_param = None
                magmom = None
            
            # transfer POSCAR and set magmom and pp/orb, and write STRU file
            if os.path.isfile(os.path.join(job, "POSCAR")):    
                stru = self.transfer_poscar(os.path.join(job, "POSCAR"))
                elements = stru.get_element(total=False, number=False)
                natoms = stru.get_natoms()

                # set magmom
                if magmom is not None:
                    if len(magmom) not in [natoms, 3*natoms]:
                        raise ValueError(f"Magmom length {len(magmom)} does not match the number of atoms {natoms}.")
                    if len(magmom) == 3 * natoms:
                        if input_param.get("nspin", 1) != 4:
                            warnings.warn(f"Warning: the magmom length {len(magmom)} is 3 times the number of atoms, but nspin is {input_param.get('nspin', 1)}, set nspin to 4.")
                            input_param["nspin"] = 4
                        magmom = [magmom[i:i+3] for i in range(0, len(magmom), 3)]
                    else:
                        if input_param.get("nspin", 1) != 2:
                            warnings.warn(f"Warning: the magmom length {len(magmom)} is equal to the number of atoms, but nspin is {input_param.get('nspin', 1)}, set nspin to 2.")
                            input_param["nspin"] = 2
                    stru.set_atommag(magmom)

                # set pp orb
                pps, orbs = self.set_pporb(elements)
                if pps is not None:
                    pp_names = self.link_pporb(job, pps)
                    stru.set_pp(pp_names)
                if orbs is not None:
                    orb_names = self.link_pporb(job, orbs)
                    stru.set_orb(orb_names)
                
                # write STRU file
                stru.write(os.path.join(job, "STRU"))
            else:
                stru, elements = None, None
            
            # set ecutwfc and basis, and write INPUT
            if input_param is not None:
                input_param.update(self.input_template)  # overwrite with template
                # only pw basis will find the recommand ecutwfc
                if input_param.get("basis_type", "pw") == "pw":
                    recommand_ecutwfc = self.find_recommand_ecutwfc(elements)
                else:
                    recommand_ecutwfc = None
                input_param = self.set_default_input(input_param, recommand_ecutwfc)
                
                WriteInput(input_param, os.path.join(job, "INPUT"))

            # transfer KPT if exists
            if os.path.isfile(os.path.join(job, "KPOINTS")):
                kpt, model = self.transfer_kpoints(os.path.join(job, "KPOINTS"))
                WriteKpt(kpt, os.path.join(job, "KPT"), model)
            
            transfer_jobs.append(job)
        
        return transfer_jobs
    
    def link_pporb(self, job_path, pporbs):
        """ Link the pseudopotentials and orbitals to the job path.
        Parameters
        ----------
        job_path : str or Path
            The path of the job.
        pporbs : List[str]
            The list of pseudopotentials and orbitals to link.
        """
        for pporb in pporbs:
            tpporb = os.path.join(job_path, os.path.basename(pporb))
            if os.path.isfile(tpporb):
                # remove the old pp/orb file if exists
                os.remove(tpporb)
            if os.path.isfile(pporb):
                os.symlink(pporb, tpporb)   
        return [os.path.basename(pporb) for pporb in pporbs]     
    
    def set_pporb(self, elements):
        pps = None
        orbs = None
        
        if self.pp_lib is not None:
            element_pp_lack = [e for e in elements if e not in self.pp_lib]
            if len(element_pp_lack) > 0:
                warnings.warn(f"Warning: the following elements do not have pseudopotentials in the library {self.pp_path}: {', '.join(element_pp_lack)}")
                
            pps = [self.pp_lib.get(e, "") for e in elements]
        if self.orb_lib is not None:
            element_orb_lack = [e for e in elements if e not in self.orb_lib]
            if len(element_orb_lack) > 0:
                warnings.warn(f"Warning: the following elements do not have orbitals in the library {self.orb_path}: {', '.join(element_orb_lack)}")
            orbs = [self.orb_lib.get(e, "") for e in elements]
        
        return pps, orbs
            
    
    def set_default_input(self, input_param, recommand_ecutwfc=None):
        """ Set the default input parameters for ABACUS INPUT.
        Parameters
        ----------
        input_param : dict
            The input parameters to set.
            
        Returns
        -------
        input_param : dict
            The input parameters with default values set.
            
        Will set the following default parameters:
            basis_type: "pw" or "lcao"
            ecutwfc: 80 Ry for pw, 100 Ry for lcao.
            ks_solver: dav_subspace for pw, genelpa for lcao
            scf_thr: 1e-8 for pw, 1e-7 for lcao
        """
        if "basis_type" not in input_param:
            input_param["basis_type"] = "pw"
        
        if "ecutwfc" not in input_param:
            if input_param["basis_type"] == "lcao":
                input_param["ecutwfc"] = 100.0
            elif recommand_ecutwfc is not None:
                input_param["ecutwfc"] = recommand_ecutwfc
            else:
                input_param["ecutwfc"] = 80.0
        
        if "ks_solver" not in input_param:
            if input_param["basis_type"] == "lcao":
                input_param["ks_solver"] = "genelpa"
            else:
                input_param["ks_solver"] = "dav_subspace"
        
        if "scf_thr" not in input_param:
            if input_param["basis_type"] == "lcao":
                input_param["scf_thr"] = 1e-7
            else:
                input_param["scf_thr"] = 1e-8
        return input_param

    def find_recommand_ecutwfc(self, elements):
        if elements is None:
            return None
        recommand_e = [self.recommand_ecutwfc.get(e, None) for e in elements]
        recommand_e_nonone = [e for e in recommand_e if e is not None]
        
        if len(recommand_e_nonone) > 0:
            print(f"Recommanded ecutwfc for elements {elements} is {recommand_e}, set to {max(recommand_e_nonone)} Ry.")
            return max(recommand_e_nonone)
        else:
            warnings.warn(f"Warning: no recommand ecutwfc found for elements {elements}")
            return None

    def transfer_poscar(self, poscar: Union[str,Path]):
        """ Transfer a VASP POSCAR file to ABACUS structure file.
        Parameters
        ----------
        poscar : str or Path
            The path of the VASP POSCAR file.
        Returns
        -------
        stru : AbacusStru
            The ABACUS structure object.
        """
        if not os.path.exists(poscar):
            raise ValueError(f"POSCAR file {poscar} does not exist.")
        stru = AbacusStru.FromDpdata(poscar, "poscar")  
        
        return stru
    
    def transfer_incar(self, incar: Union[str,Path]):
        """ Transfer a VASP INCAR file to ABACUS input parameters.
        Parameters
        ----------
        incar : str or Path
            The path of the VASP INCAR file.
        
        Returns
        -------
        input_param : dict
            The ABACUS input parameters.
            
        Supported keys:
        - SYMPREC: symmetry_prec
        - NPAR: kpar
        - NELM: scf_nmax
        - ISPIN: nspin
        - LSORBIT: lspinorb
        - LNONCOLLINEAR: noncolin
        - IBRION: calculation type (scf, md, relax, cell-relax)
        - ISIF: calculation type (relax, cell-relax, fixed_axes)
        - NSW: number of ionic steps (used to determine IBRION)
        - EDIFFG: force_thr_ev (if negative, converted to positive)
        - ISMEAR: smearing_method (gauss, fd, mp, mp2)
        - SIGMA: smearing_sigma (in eV, converted to Ry)
        - MAGMOM: magnetic moments (if nspin is 2, length should match natoms; if nspin is 4, length should be 3 times natoms)
        - LDAU, LDAUTYPE, LDAUL, LDAUU, LDAUJ: U parameters (orbital_corr, hubbard_u, onsite_radius)
        
        """
        from pymatgen.io.vasp import Incar
        vasp_param = Incar.from_file(incar)
        vasp_param = {k.upper(): vasp_param[k] for k in vasp_param}  # convert keys to uppercase
        
        input_param = {}
        
        # remove useless keys
        useless_keys = ["SYSTEM"]
        for key in useless_keys:
            vasp_param.pop(key, None)
            
        # transfer one-to-one keys
        oneone_keys = { # the value of this keys can transfer directly
            "SYMPREC": "symmetry_prec",
            "KPAR": "kpar",
            "NELM": "scf_nmax",
            "ISPIN": "nspin"
        }
        for key in oneone_keys:
            if key in vasp_param:
                input_param[oneone_keys[key]] = vasp_param.pop(key)
        
        # transfer noncollinear and SOC
        if "LSORBIT" in vasp_param or "LNONCOLLINEAR" in vasp_param:
            input_param["lspinorb"] = vasp_param.pop("LSORBIT", False)
            if vasp_param.pop("LNONCOLLINEAR", False) or input_param["lspinorb"]:
                input_param["nspin"] = 4
                input_param["noncolin"] = True   
        
        # transfer IBRION, ISIF, NSW
        nsw = vasp_param.pop("NSW", 0)
        has_isif = "ISIF" in vasp_param
        if nsw in [-1, 0]:
            ibrion = vasp_param.pop("IBRION", -1)
        else:
            ibrion = vasp_param.pop("IBRION", 0)
        if ibrion == 0:
            isif = vasp_param.pop("ISIF", 0)
        else:
            isif = vasp_param.pop("ISIF", 2)
        
        if has_isif:
            input_param["cal_force"] = 1
            input_param["cal_stress"] = 1
        input_param.update(self.transfer_incar_calculation(ibrion, isif))

        # set EDIFFG
        if "EDIFFG" in vasp_param:
            if vasp_param["EDIFFG"] < 0:
                input_param["force_thr_ev"] = -1 * vasp_param.pop("EDIFFG")
            else:
                warnings.warn(f"Warning: the EDIFFG value {vasp_param['EDIFFG']} is not supported in ABACUS, use 0.01 eV/Ang.")
                input_param["force_thr_ev"] = 0.01
                vasp_param.pop("EDIFFG", None)
            
        # transfer smearing
        input_param.update(self.transfer_incar_smear(vasp_param.pop("ISMEAR", None), 
                                                     vasp_param.pop("SIGMA", None)))

        # get magmom
        magmom = vasp_param.pop("MAGMOM", None)
        
        # transfer U
        input_param.update(self.transfer_incar_u(vasp_param.pop("LDAU", None),
                                                 vasp_param.pop("LDAUTYPE", None),
                                                 vasp_param.pop("LDAUL", None),
                                                 vasp_param.pop("LDAUU", None),
                                                 vasp_param.pop("LDAUJ", None)))

        # ignore other keys
        if len(vasp_param) > 0:
            print(f"Warning: the following keys in INCAR are not supported in ABACUS, ignored:\n\t {', '.join(vasp_param.keys())}")
        
        return input_param, magmom   
    
    def transfer_incar_calculation(self, ibrion, isif):
        """ Set the calculation type based on IBRION and ISIF values.
        Parameters
        ----------
        ibrion : int
            The IBRION value from VASP INCAR.
        isif : int
            The ISIF value from VASP INCAR.
        Returns
        -------
        input_param : dict
            The ABACUS input parameters with the calculation type set.
        """
        input_param = {}
        if ibrion == -1:
            input_param["calculation"] = "scf"
        elif ibrion == 0:
            input_param["calculation"] = "md"
        elif ibrion in [1,2]:
            if ibrion == 1:
                warnings.warn("Warning: the IBRION value 1 is not supported in ABACUS, use cg algorithm.")
            input_param["relax_method"] = "cg"
            
            if isif < 3:
                input_param["calculation"] = "relax"
            elif isif < 9:
                input_param["calculation"] = "cell-relax"
                if isif == 3: # relax cell shape and volume
                    pass
                elif isif == 4: # fix volume
                    input_param["fixed_axes"] = "volume"
                elif isif == 8: # fix shape
                    input_param["fixed_axes"] = "shape"
                else:
                    warnings.warn(f"Warning: the ISIF value {isif} is not supported in ABACUS, use ISIF=3.")
            else:
                warnings.warn(f"Warning: the ISIF value {isif} is not supported in ABACUS, use ISIF=3.")
                input_param["calculation"] = "cell-relax"
        else:
            warnings.warn(f"Warning: the IBRION value {ibrion} is not supported in ABACUS, do SCF calculation.")
            input_param["calculation"] = "scf"
        
        return input_param
    
    def transfer_incar_smear(self, ismear, sigma):
        if ismear is None:
            return {}
        input_param = {} 
        
        if ismear == 0:
            input_param["smearing_method"] = "gauss"
        elif ismear == -14:
            input_param["smearing_method"] = "fd"
        elif ismear == 1:
            input_param["smearing_method"] = "mp"
        elif ismear == 2:
            input_param["smearing_method"] = "mp2"
        else:
            warnings.warn(f"Warning: the ISMEAR value {ismear} is not supported in ABACUS, use gauss.")
            input_param["smearing_method"] = "gauss"
        
        if sigma is not None:
            input_param["smearing_sigma"] = sigma * EV2RY    

        return input_param
    
    def transfer_incar_u(self, ldau, utype, ul, uu, uj):
        if ldau is None or not ldau:
            return {}
        
        if utype == 3:
            warnings.warn("Warning: the U type 3 is not supported in ABACUS, skip the settings of U")
            return {}
        
        if ul is None or uu is None:
            warnings.warn("Warning: the U parameters are not fully set, skip the settings of U")
            return {}
        
        if uj is None:
            uj = [0 for _ in uu]
        
        uu = [uu[i]-uj[i] for i in range(len(uu))]  # convert uu to the form of U - J
        
        input_param = {
            "dft_plus_u": 1,
            "orbital_corr": " ".join([str(i) for i in ul]),
            "hubbard_u": " ".join([str(i) for i in uu]),
            "onsite_radius": 3.0
        }
        return input_param

    def transfer_kpoints(self, kpoint_file: Union[str,Path]):
        """ Transfer a VASP KPOINTS file to ABACUS KPT file.
        Parameters
        ----------
        kpoint_file : str or Path
            The path of the VASP KPOINTS file.
        Returns
        -------
        kpt : list
            The k-point setting
        """
        if not os.path.exists(kpoint_file):
            raise ValueError(f"KPOINTS file {kpoint_file} does not exist.")
        
        with open(kpoint_file, "r") as f: lines = f.readlines()
        
        assert len(lines) >= 5, "KPOINTS file should have at least 5 lines."
        
        nks = int(lines[1].strip())
        model = lines[2].strip().lower()
        if nks == 0:
            if model.startswith("g") or model.startswith("m"):
            # Gamma or Monkhorst-Pack k-point
                kpt = [int(i) for i in lines[3].strip().split()] + [float(i) for i in lines[4].strip().split()[:3]]
                model = "gamma" if model.startswith("g") else "mp"
            else:
                raise ValueError(f"The automatic mesh generation method {model} is not supported, should be 'Gamma' or 'MP'.")
        elif nks > 0:
            if model.startswith("l"):
                # line model
                if lines[3].strip().lower().startswith("c") or lines[3].strip().lower().startswith("k"):
                    model = "line_cartesian"
                else:
                    model = "line"
                    
                k_pairs = []
                pair_lines = [line.strip() for line in lines[4:] if line.strip()]
                assert len(pair_lines) % 2 == 0, "Line k-point pairs should be in pairs."
                for i in range(0, len(pair_lines), 2):
                    k1_lines = pair_lines[i].split()
                    k2_lines = pair_lines[i+1].split()
                    k1 = [float(k) for k in k1_lines[:3]]
                    k2 = [float(k) for k in k2_lines[:3]]
                    if len(k1_lines) > 3:
                        k1.append("# " + " ".join(k1_lines[3:]))
                    if len(k2_lines) > 3:
                        k2.append("# " + " ".join(k2_lines[3:]))
                    k_pairs.append([k1, k2])
                
                kpt = []
                pre_k2_coord = None
                pre_k2 = None
                for k1, k2 in k_pairs:
                    k1_coord = "_".join([f"{k:.6f}" for k in k1[:3]])
                    k2_coord = "_".join([f"{k:.6f}" for k in k2[:3]])
                    
                    if pre_k2_coord is not None and pre_k2_coord != k1_coord:
                        pre_k2.insert(3, 1)
                        kpt.append(pre_k2)
                    k1.insert(3, nks)
                    kpt.append(k1)
                    
                    pre_k2 = k2
                    pre_k2_coord = k2_coord
                pre_k2.insert(3, 1)
                kpt.append(pre_k2)
            else:
                if model.startswith("c") or model.startswith("k"):
                    model = "cartesian"
                else:
                    model = "direct"
                kpt = []
                for line in lines[3:]:
                    if line.strip() == "":
                        continue
                    kpt.append([float(k) for k in line.strip().split()[:4]])
        else:
            raise ValueError(f"Invalid k-point setting in KPOINTS file: {kpoint_file}. The number of k-points should be non-negative.")
        
        return kpt, model


import os
from .comm_class import OutputSet,ConfigSet,ConfigSetGithub
from . import comm_pmetrics,comm_func

from dp.launching.report import Report,AutoReportElement,ReportSection,ChartReportElement

def gen_report_param(params,show_param=None):
    """
    generate a report section for the parameters
    """
    # get all member of params
    report_member = "Input Parameters:\n"
    if show_param is None:
        config_members = list(ConfigSet.__fields__.keys()) + list(ConfigSetGithub.__fields__.keys()) + list(OutputSet.__fields__.keys())

        
        for im in params.__fields__.keys():
            if im not in config_members:
                report_member += f"{im}={getattr(params,im)}\n"
    else:
        for im in show_param:
            if im in params.__fields__.keys():
                report_member += f"{im}={getattr(params,im)}\n"
    
    report_member += "\n"
    return report_member

def gen_report(opts,logs,work_path,output_path,abacustest_param):
    param_c = gen_report_param(opts)
    with open(os.path.join(output_path,"ABACUSTEST_INPUT_PARAM.txt"),"w") as f:
        f.write(param_c)
    
    reports = [
        ReportSection(title="INPUT PARAMETER",
                      elements=[AutoReportElement(title='', path="ABACUSTEST_INPUT_PARAM.txt", description="")])
    ]
    if abacustest_param is not None:
        reports += comm_pmetrics.produce_metrics_superMetrics_reports(abacustest_param,work_path,output_path)
    logfname = "output.log"
    logs.write(os.path.join(output_path,logfname))
    log_section = ReportSection(title="",
                              elements=[AutoReportElement(title='', path=logfname, description="")])
    reports.append(log_section)
    if reports:
        report = Report(title="abacus test report",
                        sections=reports,
                        description="a report of abacustest")
        report.save(output_path)
    # move results to output_path
    if abacustest_param is not None:
        comm_func.move_results_to_output(work_path,output_path,abacustest_param.get("save_path","results"))
    else:
        os.system(f"mv {work_path}/* {output_path}/")
    #comm_func.pack_results(output_path,abacustest_param.get("save_path","results"))


import glob,os,json
import numpy as np
from abacustest.lib_collectdata.collectdata import RESULT
from . import comm_plot,comm

def shift_data(data_list, base_index=0):
    # shift the data to the base_index
    # data_list should a list of float or list of list of float
    if not isinstance(data_list,list):
        return None
    if base_index < 0 or base_index >= len(data_list):
        return None
    
    if isinstance(data_list[base_index],list):
        # it is a list of list of float
        new_list = []
        if None in data_list[base_index]:
            return None
        for i in range(len(data_list)):
            if len(data_list[i]) != len(data_list[base_index]):
                new_list.append(None)
            else:
                new_list.append([data_list[i][j] - data_list[base_index][j] for j in range(len(data_list[i]))])
    else:
        if data_list[base_index] is None:
            return None
        new_list = [data_list[i] - data_list[base_index] for i in range(len(data_list))]
    return new_list

class PostConv:
    """
    Used to do the postprocess of convergence test of ecutwfc and kspacing.
    
    It will firstly collect the metrics of the jobs, and save the results to metrics.json file.
    If the metric_file is defined, it will also read the results in file, and merge the results. This file should have the same format as metrics.json.
    
    The it will rearrnge the results to plotdata, and plot the results.
    
    For default, energy/force/stress/atom_mag/mag_force/band_gap will be plotted.
    
    Args:
        test_key: str, the key of the test, ecutwfc or kspacing. It should be a parameter name of INPUT. Special case is kpt, which will be used to plot the k-point mesh.
        jobs: list of str, the list of the job folders.
        metric_file: str, the file of the metrics. or a liat of metrics files.
        extra_y: dict, the extra y to plot. The key is the metric name, and the value is the y title.
        job_type: str, abacus or qe or vasp, the type of the job.
        shift_data: max/min/False, whether to shift the data. Default is None.
            None: will shift the data to VALUE with max ecutwfc or min kspacing.
            max: will shift the data to the max test_key.
            min: will shift the data to the min test_key
            False: will not shift the data.
    """
    def __init__(self, test_key="ecutwfc",jobs=None, metric_file=None, extra_y=None, job_type="abacus",shift_data=None,x_name=None):
        self.test_key = test_key
        self.jobs = jobs
        self.metric_file = None
        if metric_file is not None:
            if isinstance(metric_file,str):
                self.metric_file = [metric_file]
            elif isinstance(metric_file,list):
                self.metric_file = metric_file
                
        self.extra_y = [extra_y] if isinstance(extra_y,str) else extra_y
        self.job_type = job_type

        self.plot_keys = {"energy_per_atom":"Energy (meV/atom)", 
                          "force":"Force (eV/$\mathrm{\AA}$)", 
                          "stress":"Stress (kbar)", 
                          "atom_mag":"Atomic maganetic moment ($\mathrm{\mu}$B)", 
                          "ds_mag_force": "Maganetic force ($\mathrm{\mu}$B/$\mathrm{\AA}$)",
                          "band_gap":"Band gap (eV)"}
        # key is the metric name, value is the y title
        
        if self.extra_y is not None:
            for iey,iey_unit in self.extra_y.items():
                self.plot_keys[iey] = iey_unit
        
        self.shift_idx = None  # shift the energy to value of shift_idx (0 means the first value, -1 means the last value)
        if shift_data == "max":
            self.shift_idx = -1
        elif shift_data == "min":
            self.shift_idx = 0
        elif shift_data == None:
            if self.test_key in ["ecutwfc", "kpt"]:
                self.shift_idx = -1
            elif self.test_key == "kspacing":
                self.shift_idx = 0
        
        if x_name is not None:
            self.x_name = x_name.capitalize()
        else:
            if test_key == "ecutwfc":
                self.x_name = "Ecutwfc (Ry)"
            elif test_key == "kspacing":
                self.x_name = "Kspacing (1/bohr)"
            elif test_key == "kpt":
                self.x_name = "K-point mesh"
            else:
                self.x_name = test_key.capitalize()

    def run(self):
        metrics_jobs = None
        metrics_file = []
        if self.jobs is not None:
            metrics_jobs = self._collect_metrics()
            json.dump(metrics_jobs, open("metrics.json", "w"), indent=4)
        if self.metric_file is not None:
            metrics_file = [json.load(open(i)) for i in self.metric_file if os.path.isfile(i)]
        
        if metrics_jobs is None and len(metrics_file)==0:
            print("No metrics are found.")
            return
        
        plotdata = self.rearrage_data(metrics_jobs, metrics_file)
        plotdata = self.sort_plotdata(plotdata)
        fnames = self.plot_data(plotdata)
        
        # write a supermetrics.json
        json.dump(plotdata, open("result.json", "w"), indent=4)
        if fnames:
            json.dump({k:{"type": "image", "file": k}  for k in fnames}, open("supermetrics.json","w"),indent=4)
    
    def _collect_metrics(self):
        jobs = self.jobs
        test_key = self.test_key
        job_type = self.job_type
        
        allmetrics = {}
        for job in jobs:
                for ijob in glob.glob(os.path.join(job,"*")):
                    if os.path.isdir(ijob):
                        results = RESULT(path=ijob,fmt=job_type)
                        if test_key == "kpt":
                            test_key_v = results["kpt"]
                        else:
                            if job_type == "abacus":
                                test_key_v = results["INPUT"].get(test_key,None)
                            else:
                                test_key_v = results[test_key]
                                
                        allmetrics[ijob] = {
                            test_key: test_key_v,
                            "energy_per_atom": results["energy_per_atom"],
                            "energy": results["energy"],
                            "natom": results["natom"],
                            "normal_end": results["normal_end"],
                            "denergy_last": results["denergy_last"],
                            "drho_last": results["drho_last"],
                            "converge": results["converge"],
                            "band_gap": results["band_gap"],
                            "scf_steps": results["scf_steps"],
                            "total_time": results["total_time"],
                            "scf_time": results["scf_time"],
                            "force": results["force"],
                            "stress": results["stress"],
                            "ds_mag_force": results["ds_mag_force"],
                            "version": results["version"],
                            "ks_solver": None if not isinstance(results["INPUT"],dict) else results["INPUT"].get("ks_solver",None),
                            "scf_time/step": None if not results["scf_time"] or not results["scf_steps"] else results["scf_time"] / results["scf_steps"]
                        }
                        for ikey in self.plot_keys:
                            if ikey not in allmetrics[ijob]:
                                allmetrics[ijob][ikey] = results[ikey]         
        return allmetrics

    def get_xkey_v(self, iv):
        # we need deal with the x_key value to transfer it to a float or int or string
        x_key_v = iv.get(self.test_key,None)
        if x_key_v is None:
            return None
        
        if self.test_key == "kspacing":
            kspacing = x_key_v
            if isinstance(kspacing,str):
                try:
                    ksplit = [float(ik) for ik in kspacing.split()]
                    if len(ksplit) == 1:
                        kspacing = ksplit[0]
                    elif len(ksplit) == 3:
                        if ksplit[0] == ksplit[1] == ksplit[2]:
                            kspacing = ksplit[0]
                except:
                    return kspacing
            return kspacing
        elif isinstance(x_key_v,str):
            x_key_v_split = x_key_v.split()
            if len(x_key_v_split) == 1:
                try:
                    return float(x_key_v)
                except:
                    return x_key_v
            elif len(x_key_v_split) > 1:
                try:
                    return tuple([float(i) for i in x_key_v_split])
                except:
                    return x_key_v

        return x_key_v
                
    def get_ykey_v(self, iv, y_key):
        y_key_v = iv.get(y_key,None)
        if y_key_v is None:
            return None
        
        if y_key in ["energy","energy_per_atom"]:
            # here will actually use energy_per_atom in unit meV/atom
            if "energy_per_atom" in iv:
                y_key_v = iv["energy_per_atom"]
            else:
                e = iv.get("energy",None)
                na = iv.get("natom",None)
                if e is not None and na is not None:
                    y_key_v = e / na
                elif e is not None and na is None:
                    y_key_v = e
                    self.plot_keys[y_key] = "meV"
                else:
                    y_key_v = None
            if y_key_v is not None:
                y_key_v *= 1000 # transfer to meV/atom
        elif y_key == "mag_force":
            return iv.get("ds_mag_force",None)
        
        # if y_key_v is a list of list, we should transfer it to one list
        if isinstance(y_key_v,list):
            y_key_v = np.array(y_key_v).flatten().tolist()
                
        return y_key_v   
    
    def rearrage_data(self, metrics_jobs, metrics_file):
        # transfer the metrics to a dict with key is example name and value is a dict of subfolder, x_key, and y_keys
        '''
        {
            "example_name": {
                "subfolder": [],
                "x_key": [],  # x_key should be a list of float or int or string
                "y_key1": [], # y_key should be a list of float or int or list, element of y_key is matched with x_key
                "y_key2": [],
                ...
            }
        }
        '''
        plotdata = {}
        x_key = self.test_key
        for allmetrics in [metrics_jobs] + metrics_file:
            if allmetrics is None:
                continue
            
            for i, iv in allmetrics.items():
                x_value = self.get_xkey_v(iv)
                if x_value is None: # do not save the data if the x_key is None
                    continue
                example_name = os.path.dirname(i.rstrip("/"))
                if example_name not in plotdata:
                    plotdata[example_name] = {
                        "subfolder": [],
                        x_key: [],
                    }
                    for ik in self.plot_keys:
                        plotdata[example_name][ik] = []

                if x_value in plotdata[example_name][x_key]: # if one x_value is already in the plotdata, skip it. For case value from jobs and file are the same.
                    continue
                plotdata[example_name]["subfolder"].append(os.path.basename(i.rstrip("/")))
                plotdata[example_name][x_key].append(x_value)
                for ik in self.plot_keys:
                    plotdata[example_name][ik].append(self.get_ykey_v(iv,ik))   
        return plotdata
    
    def sort_kpt(self, kpts):
        # kpts is a list of k-point meshs, such as [[2,2,2],[3,3,3],[4,4,4]]
        # we need sort the kpts by the number of k-points, and then by the k-point mesh
        # the number of k-points is the product of the three elements
        kpoints = [ np.prod(kpt) for kpt in kpts ]
        sort_idx = sorted(range(len(kpoints)), key=lambda k: (kpoints[k], kpts[k]))
        
        # if 3 k-points are the same, then return the first one
        if all([len(set(k)) == 1 for k in kpts]):
            kpts = [i[0] for i in kpts]
        else:
            kpts = ["_".join([str(j) for j in i]) for i in kpts]
        kpts = [kpts[i] for i in sort_idx]
        return sort_idx, kpts
        
    
    def sort_plotdata(self, plotdata):
        # we need sort the plotdata by x_key
        # and if shift_idx is not False, we need shift the data to the shift_idx
        new_plotdata = {}
        for ik,iv in plotdata.items():
            
            if self.test_key == "kpt":
                # if the test_key is kpt, we need sort the kpts
                sort_idx, x_values = self.sort_kpt(iv[self.test_key])
            else:
                x_values = iv[self.test_key]
                sort_idx = sorted(range(len(x_values)), key=lambda k: x_values[k])
                x_values = [x_values[i] for i in sort_idx]
                
            new_plotdata[ik] = {
                "subfolder": [iv["subfolder"][i] for i in sort_idx],
                self.test_key: x_values
            }
            for ikey in self.plot_keys:
                if self.shift_idx is None or ikey in ["band_gap"]:
                    new_plotdata[ik][ikey] = [iv[ikey][i] for i in sort_idx]
                else:
                    new_plotdata[ik][ikey] = shift_data(iv[ikey],base_index=sort_idx[self.shift_idx])
        return new_plotdata
    
    def plot_data(self, plotdata):
        allexamples = list(plotdata.keys())
        allexamples.sort()
        fnames = []
        ykeys = list(self.plot_keys.keys())
        ykeys.sort()
        
        for ik in allexamples:
            
            x = plotdata[ik][self.test_key]
            
            for ykey in ykeys:
                fname = ik.replace("/","_") + f"-{ykey}.png"
                y = plotdata[ik][ykey]
                if y is None:
                    continue
                
                ix, iy = comm.clean_none_list(x,y)
                if len(ix) == 0:
                    continue
                if isinstance(ix[0],(tuple,list)):
                    ix = [str(i) for i in ix]
                
                # if iy is a list of list, we should rotate it
                if isinstance(iy[0],list):
                    iys = np.array(iy).T
                else:
                    iys = [iy]

                title = f"{ik}({ykey})"
                ytitle = self.plot_keys[ykey]
                if self.shift_idx is None or ykey in ["band_gap"]:
                    ytitle = ytitle.capitalize()
                else:
                    ytitle = "Relative " + ytitle
                        
                comm_plot.plot_line_point(ix, iys,xtitle=self.x_name,ytitle=ytitle,title=title, 
                                      fname=fname,figsize=(8,6),fontsize=16,grid=True)
                fnames.append(fname)
        
        return fnames        
                
                
            


#!/usr/bin/env python
from datetime import datetime
def _init():
    global _global_dict
    _global_dict = {"START_TIME":datetime.now()}

def set_value(key,value):
    # check if _global_dict has been initialized
    if "_global_dict" not in globals():
        _init()
    
    if key in _global_dict:
        print("WARNING: key %s has been used, the value will be modified from '" % key,_global_dict[key],"' to '",value,"'")
    _global_dict[key] = value

def get_value(key,defValue=None):
    if "_global_dict" not in globals():
        _init()
    return _global_dict.get(key,defValue)







import traceback,json,sys
from dp.launching.typing.basic import BaseModel,String
from dp.launching.typing import Field
import os



from . import (comm_class,
               comm_func,
               comm_class_exampleSource,
               comm_report) 

'''
class Models(String, Enum):
    model00 = " "
    model0 = "000-abacus-branch"
    model1 = "001-abacus-dailytest"
    #model3 = 
    model4 = "004-lcao004_unstable_elpa"
    model4_1 = "004-lcao004_unstable_scalapack"
    model5 = "005-finite_diff_stress"
    model5_1 = "005-finite_diff_force"  
    model6 = "006_eos"
    model7 = "007_bandstru"
    model7_1 = "007_bandstru_abacusVSvasp"
    model8 = "008_dos"
    model10 = "010_LcaoVSPw"
    model11 = "011_abacusVSvasp_eos"
    model11_1 = "011_LcaoVSPw_eos"
    
class MyModel(BaseModel):
    my_model: Models = Field(
        title="Model",
        default="",description="The model you want to use. If you have entered the dataset information in the previous step, here will be ignored. DOCs: https://dptechnology.feishu.cn/docx/I7NKdmURHosJSMxn7NCc5RsEnxb")
'''
class NewSetting(BaseModel):
    predft_command: String = Field(default="",
                            title="New Predft Command",
                            description="If you don't want to use the command from the previously uploaded setting file, you can fill in the new command here",)
    rundft_image: String = Field(default="",
                          title="New Rundft Image",
                          description="If you don't want to use the image from the previously uploaded setting file, you can fill in the new command here.",)
    rundft_command: String = Field(default="",
                            title="New Rundft Command",
                            description="If you don't want to use the command from the previously uploaded setting file, you can fill in the new command here",)
    rundft_machine: String = Field(default="",
                            title="New Rundft Machine",
                            description="If you don't want to use the machine from the previously uploaded setting file, you can fill in the new command here.",)
    postdft_command: String = Field(default="",
                            title="New Postdft Command",
                            description="If you don't want to use the command from the previously uploaded setting file, you can fill in the new command here.",)

'''
group1 = ui.Group("if_use_reuse_dataset","test")

@group1
@ui.Visible(MyModel,("my_model"),Equal,("do not show this group"))
class ReuseDataset(BaseModel):
    reuse_dataset: DataSet = Field(title=None,
                             default="launching+datasets://reuse.abacustest@latest",
                            description="the reuse dataset")
'''                            
        
class ReuseModel(
    #comm_class.ConfigSetGithub,
    #comm_class.TrackingSet,
    NewSetting,
    #comm_class_exampleSource.ScriptExampleSet,
    #comm_class_exampleSource.ScriptSourceSet,
    #comm_class_exampleSource.ScriptDatasetSet,
    #comm_class_exampleSource.ExampleSet,
    comm_class_exampleSource.ExampleSourceSet,
    #comm_class_exampleSource.DatasetSet,
    #MyModel,
    #ReuseDataset,
                    comm_class.OutputSet,
                    comm_class.ConfigSet,
                    BaseModel):
    ...  

def ReuseModelRunner(opts:ReuseModel) -> int:
    try:
        logs = comm_class.myLog()

        paths = comm_func.create_path(str(opts.IO_output_path))
        output_path = paths["output_path"]
        work_path = paths["work_path"]
        download_path = paths["download_path"]

        logs.iprint("read source setting ...")
        datas = comm_class_exampleSource.read_source(opts,work_path,download_path,logs.iprint)
        
        setting_file = os.path.join(work_path,"setting.json")
        if not os.path.exists(setting_file):
            logs.iprint("\nERROR: Please supply the setting information!!!!")
            return 1
        try:
            setting = json.load(open(setting_file))
        except:
            traceback.print_exc()
            return 1
        
        #read setting
        allparams = {"config": comm_func.read_config(opts)}
        for k,v in setting.items():
            if k == "config":
                for ik,iv in v.items():
                    allparams["config"][ik] = iv
            #elif k in ["ABBREVIATION","save_path","run_dft","post_dft","report","dataset_info","upload_datahub","upload_tracking"]:
            else:
                allparams[k] = v
                
        # modify the command and image
        if opts.predft_command != None and opts.predft_command.strip() != "":
            new_predft_command = opts.predft_command.strip()
        else:
            new_predft_command = None
        if opts.rundft_image != None and opts.rundft_image.strip() != "":
            new_rundft_image = opts.rundft_image.strip()
        else:
            new_rundft_image = None
        if opts.rundft_command != None and opts.rundft_command.strip() != "":
            new_rundft_command = opts.rundft_command.strip()
        else:
            new_rundft_command = None
        if opts.rundft_machine != None and opts.rundft_machine.strip() != "":
            new_rundft_machine = opts.rundft_machine.strip()
        else:
            new_rundft_machine = None
        if opts.postdft_command != None and opts.postdft_command.strip() != "":
            new_postdft_command = opts.postdft_command.strip()
        else:
            new_postdft_command = None
        
        if "pre_dft" in allparams:
            if new_predft_command:
                allparams["pre_dft"]["command"] = new_predft_command
        
        if "run_dft" in allparams:
            if isinstance(allparams["run_dft"],list):
                for idx in range(len(allparams["run_dft"])):
                    if new_rundft_image:
                        allparams["run_dft"][idx]["image"] = new_rundft_image
                    if new_rundft_command:
                        allparams["run_dft"][idx]["command"] = new_rundft_command
                    if new_rundft_machine and "bohrium" in allparams["run_dft"][idx] and isinstance(allparams["run_dft"][idx]["bohrium"],dict):
                        allparams["run_dft"][idx]["bohrium"]["scass_type"] = new_rundft_machine
            elif isinstance(allparams["run_dft"],dict):
                if new_rundft_image:
                    allparams["run_dft"]["image"] = new_rundft_image
                if new_rundft_command:
                    allparams["run_dft"]["command"] = new_rundft_command
                if new_rundft_machine and "bohrium" in allparams["run_dft"] and isinstance(allparams["run_dft"]["bohrium"],dict):
                    allparams["run_dft"]["bohrium"]["scass_type"] = new_rundft_machine
        
        if "post_dft" in allparams:
            if new_postdft_command:
                allparams["post_dft"]["command"] = new_postdft_command
     
        #execut
        stdout,stderr = comm_func.exec_abacustest(allparams,work_path)
        logs.iprint(f"{stdout}\n{stderr}\nrun abacustest over!\n")
        
        comm_report.gen_report(opts,logs,work_path,output_path,allparams)
        
    except:
        traceback.print_exc()
        return 1

    return 0



from dp.launching.typing.basic import BaseModel
import os
from dp.launching.report import Report,AutoReportElement,ReportSection,ChartReportElement

from . import (comm_class,
               comm_func,
               comm_pmetrics,
               comm_class_exampleSource,
               comm_class_rundft,
               comm_class_postdft,
               comm_class_metrics,
               readsetting)


class NormalModel(comm_class.TrackingSet,
                  comm_class_metrics.MetricsSet,
                  comm_class_postdft.PostdftImageSet,
                  comm_class_rundft.RundftImageSet,
                  comm_class_rundft.RundftCommandSet,
                  comm_class_rundft.RundftGroupSizeSet,
                  comm_class_exampleSource.RundftExampleSet,
                  comm_class_exampleSource.RundftExampleSourceSet,
                  comm_class.OutputSet,
                  comm_class.ConfigSet,
                  BaseModel):
    ...  

class NormalDatasetsModel(comm_class.TrackingSet,
                  comm_class_metrics.MetricsSet,
                  comm_class_postdft.PostdftImageSet,
                  comm_class_rundft.RundftImageSet,
                  comm_class_rundft.RundftCommandSet,
                  comm_class_rundft.RundftGroupSizeSet,
                  comm_class_exampleSource.RundftExampleSet,
                  comm_class_exampleSource.DatasetSet,
                  comm_class.OutputSet,
                  comm_class.ConfigSet,
                  BaseModel):
    ...  

def NormalModelRunner(opts) -> int:
    logs = comm_class.myLog()

    paths = comm_func.create_path(str(opts.IO_output_path))
    output_path = paths["output_path"]
    work_path = paths["work_path"]
    download_path = paths["download_path"]

    allparams = readsetting.ReadSetting(logs,opts,work_path,download_path)
    if allparams == None:
        return 1

    stdout,stderr = comm_func.exec_abacustest(allparams,work_path)
    logs.iprint(f"{stdout}\n{stderr}\nrun abacustest over!\n")
    reports = comm_pmetrics.produce_metrics_superMetrics_reports(allparams,work_path,output_path)

    logfname = "output.log"
    logs.write(os.path.join(str(opts.IO_output_path),logfname))
    log_section = ReportSection(title="",
                              elements=[AutoReportElement(title='', path=logfname, description="")])
    reports.append(log_section)

    if reports:
        report = Report(title="abacus test report",
                        sections=reports,
                        description="a report of abacustest")
        report.save(output_path)

    #move results to output_path
    comm_func.move_results_to_output(work_path,output_path,allparams.get("save_path","results"))
    comm_func.pack_results(output_path,allparams.get("save_path","results"))
    
    
    return 0



#!/usr/bin/env python3
import os,sys,argparse
sys.path.append(os.path.split(__file__)[0])
from .myflow import flow,globV,comm
from dflow.python import upload_packages
#import numpy,pandas
#upload_packages.append(numpy.__path__[0])
#upload_packages.append(pandas.__path__[0])
#upload_packages.append(pymatgen.__path__[0])
#upload_packages.append(os.path.join(os.path.split(__file__)[0],'myflow'))
#upload_packages.append(os.path.join(os.path.split(__file__)[0],'lib_collectdata'))
upload_packages.append(os.path.split(__file__)[0])

def AbacusTestArgs(parser):
    parser.description = "This script is used to run a testing"
    parser.add_argument('-p', '--param', type=str, default="job.json",help='the job setting file, default is job.json')
    parser.add_argument('-s', '--save', type=str, default=None,help='the folder where the results will be put in, default: result')
    parser.add_argument('--override', type=int, default=1,help="when the save folder exists, if override it. 0: no, 1: yes. ")
    parser.add_argument('--debug',nargs='?',type=int, const=1, default=0,help="debug mode for dflow")
    parser.add_argument('--download',default=1,type=int, help="if wait the finish of flow and download the results, 0: no, 1: yes. Default is 1")
    return parser

def AbacusTestCheckStatusArgs(parser):
    parser.description = "Check the status of the dflow job"
    parser.add_argument('-p', '--param', type=str, default="job.json",help='the file for bohrium account information, default is "job.json"')
    parser.add_argument("job_id", help="the job id of dflow")
    return parser

def AbacusTestDownloadArgs(parser):
    parser.description = "Download the results of the dflow job"
    parser.add_argument("job_id", help="the job id of your workflow")
    parser.add_argument('-p', '--param', type=str, default="job.json",help='the file for bohrium account information, default is "job.json"')
    parser.add_argument('-s', '--save', type=str, default=None,help='the folder where the results will be put in, default: result')
    return parser

def abacustest(param):
    globV._init()
    flow.RunJobs(param)
    comm.printinfo("\nAll jobs are finished!!!")

def checkstatus(param):
    globV._init()
    status = flow.CheckStatus(param)
    print(status)

def downloadflow(param):
    globV._init()
    flow.DownloadFlow(param)

def main():
    parser = argparse.ArgumentParser(description="abacustest")
    subparser = parser.add_subparsers(dest="command")
    AbacusTestArgs(subparser.add_parser("submit",help="Submit a workflow"))
    AbacusTestArgs(subparser.add_parser("mlops-submit",help="Just submit a workflow and will not wait for the result"))
    AbacusTestCheckStatusArgs(subparser.add_parser("status", help="Check the status of the dflow job"))
    parser = parser.parse_args()
    if parser.command == "submit":
        abacustest(parser)
    elif parser.command == "status":
        checkstatus(parser)
    elif parser.command == "download":
        downloadflow(parser)

if __name__ == '__main__':
    main()



import traceback,json,sys,glob
from dp.launching.typing.basic import BaseModel,String
from dp.launching.typing import Field
import os

from abacustest.lib_model.model_005_Phonon import PreparePhono,PostprocessPhonon
from abacustest.constant import RECOMMAND_IMAGE

from . import (comm_class,
               comm_func,
               comm_report,
               comm_class_exampleSource,
               readsetting) 


class Images(BaseModel):
    abacus_image: String = Field(default=RECOMMAND_IMAGE,
                          title="abacus Image",
                          description="If you do not want to use the default image, please enter the new image here.",)
    abacus_command: String = Field(default="OMP_NUM_THREADS=1 mpirun -np 16 abacus | tee out.log",
                            title="abacus Command",
                            description="The command to run abacus",)
    abacus_machine: String = Field(default="c32_m64_cpu",
                                   title="abacus Machine",
                                    description="The machine to run abacus",)
    phonopy_setting: String = Field(default="setting.conf",
                                   title="Phonopy Setting",
                                    description="the phonopy setting file name, and each example should have this file and containing DIM/BAND",)

class PhononModel(
    Images,
    comm_class_exampleSource.ExampleSourceSet,
                    comm_class.OutputSet,
                    comm_class.ConfigSet,
                    BaseModel):
    ...      
    
def PhononModelRunner(opts:PhononModel) -> int:   
    try:
        logs = comm_class.myLog()
        paths = comm_func.create_path(str(opts.IO_output_path))
        output_path = paths["output_path"]
        work_path = paths["work_path"]
        download_path = paths["download_path"]
        cwd = os.getcwd()
        logs.iprint("read source setting ...")
        datas = comm_class_exampleSource.read_source(opts,work_path,download_path,logs.iprint)
        allparams = {"config": comm_func.read_config(opts),
                     "save_path": ".",
                     "bohrium_group_name": "Phonopy"}
        
        examples = datas.get("all_files",[])
        if len(examples) == 0:
            return 1
        os.chdir(work_path)
        setting = PreparePhono(examples,opts.abacus_command,opts.abacus_image,opts.abacus_machine,"phonopy",opts.phonopy_setting)
        if len(setting) == 0:
            return 1
        
        # merge setting and config
        allparams.update(setting)

        #exec abacustest
        stdout,stderr = comm_func.exec_abacustest(allparams,work_path)
        
        #postprocess
        os.chdir(work_path)
        PostprocessPhonon(examples,"phonopy",opts.phonopy_setting)
        os.chdir(cwd)

        logs.iprint(f"{stdout}\n{stderr}\nrun abacustest over!\n")
        
        # produce metrics and superMetrics reports
        comm_report.gen_report(opts,logs,work_path,output_path,allparams)
    except:
        traceback.print_exc()
        return 1

    return 0


from ..model import Model
import os, glob, json
from . import comm,comm_conv,comm_plot
from abacustest.lib_collectdata.collectdata import RESULT
from abacustest.constant import RECOMMAND_IMAGE

SETTING_TMP = {
    "save_path": "results",
    "bohrium_group_name": "convergence-test",
    "prepare": {"example_template": [], "mix_input": {}},
    "run_dft": {
        "command": "OMP_NUM_THREADS=1 mpirun -n 16 abacus | tee out.log",
        "image": RECOMMAND_IMAGE,
        "bohrium": {
            "scass_type": "c32_m64_cpu",
            "job_type": "container",
            "platform": "ali",
        },
    },
    "post_dft": {
        "image": "registry.dp.tech/dptech/abacustest:latest",
        "metrics": {
            "dft_type": "abacus",
            "metrics_name": [
                "normal_end",
                "converge",
                "natom",
                "scf_steps",
                "total_time",
                "scf_time",
                "energy",
                "energy_per_atom",
                "force",
                "stress",
                "version",
                {
                    "scf_time/step": "{scf_time}/{scf_steps}",
                    "ks_solver": "{INPUT}['ks_solver']",
                    "ecutwfc": "{INPUT}['ecutwfc']",
                }
            ],
            "save_file": "metrics.json"
        },
    },
}


class ConvEcutwfc(Model):
    @staticmethod
    def description():  # type: ignore
        return "Do a convergence test"

    @staticmethod
    def model_name():  # type: ignore
        return "conv"

    @staticmethod
    def prepare_args(parser):
        parser.add_argument(
            "-j",
            "--jobs",
            help="the path of jobs",
            action="extend",
            nargs="*",
        )
        parser.add_argument(
            "-k",
            "--key",
            type=str,
            help="the key name to be tested, should be a INPUT para. Such as ecutwfc, kspacing, etc.",
        )
        parser.add_argument(
            "-v",
            "--value",
            help="the value to be tested",
            action="extend",
            nargs="*",
        )

    def run_prepare(self, params):
        real_jobs = comm.get_job_list(params.jobs)
        if len(real_jobs) == 0:
            print("No jobs found.")
            return
        print("The jobs are:", ", ".join(real_jobs))
        setting = SETTING_TMP
        setting["prepare"]["example_template"] = real_jobs
        setting["prepare"]["mix_input"][params.key] = params.value
        setting["post_dft"]["metrics"]["metrics_name"][-1][params.key] = f"{{INPUT}}['{params.key}']"
        
        comm.dump_setting(setting)
        comm.doc_after_prepare(f"convergence test of {params.key}", real_jobs, ["setting.json"])

    
    @staticmethod
    def postprocess_args(parser):
        #group = parser.add_mutually_exclusive_group()
        parser.add_argument(
            "-k",
            "--key",
            help="the key name to be tested, should be a INPUT para. Such as ecutwfc, kspacing, etc.",
        )
        parser.add_argument(
            "-j",
            "--jobs",
            help="the path of jobs. There should have several subfolders with different ecutwfc for each job.",
            action="extend",
            nargs="*",
        )
        parser.add_argument(
            "-m",
            "--metric",
            default=None,
            help="the metrics.json file. Key is the examplename/subfolder and value is a dict of metrics value containing kspacing, energy_per_atom (or energy and natom).",
        )
        parser.add_argument(
            "-e",
            "--extra",
            nargs="*",
            action="extend",
            help="the extra metrics that need to be ploted. The format is \"metrics_name:title of metrics\"",
        )
        
    def run_postprocess(self, params):
        extra_y = None
        if params.extra:
            extra_y = {}
            for e in params.extra:
                e = e.split(":")
                extra_y[e[0]] = e[1].strip()
        comm_conv.PostConv(test_key=params.key,jobs=params.jobs,metric_file=params.metric,extra_y=extra_y).run()


import traceback,json,sys,os
from dp.launching.typing.basic import BaseModel,String,Float,Int,Boolean,Set
from dp.launching.typing import Field
from enum import Enum
from abacustest.constant import RECOMMAND_IMAGE

from . import (comm_class,
               comm_func,
               comm_report,
               comm_class_exampleSource) 
from abacustest.lib_model import model_007_FDStress as FDStress


class ComponentEnum(String, Enum):
    comp11 = '11'
    comp12 = '12'
    comp13 = '13'
    comp21 = '21'
    comp22 = '22'
    comp23 = '23'
    comp31 = '31'
    comp32 = '32'
    comp33 = '33'

class NewSetting(BaseModel):
    fd_step: Float = Field(default=0.0001,
                            title="FD Step (ratio of cell deformation)",
                            description="The finite difference step size. ",)
    
    fd_number: Int = Field(default=5,
                           titile= "FD Number",
                            description="The number of finite difference steps. Will calculate extra 2*Fd_Number structures.",)

    stress_comp: Set[ComponentEnum] = Field(default=("11","12","13","22","23","33"),title="Stress Components",description="The stress components to test.",)
    
    abacus_image: String = Field(default=RECOMMAND_IMAGE,
                          title="Abacus Image",
                          description="The image to run abaucs.",)
    abacus_command: String = Field(default="OMP_NUM_THREADS=1 mpirun -np 16 abacus | tee out.log",
                            title="Abacus Command",
                            description="The command to execute abacus",)
    bohrium_machine: String = Field(default="c32_m64_cpu",
                            title="Bohrium Machine",
                            description="The bohrium machine type to run abacus",)


class FDStressModel(
    #comm_class.ConfigSetGithub,
    #comm_class.TrackingSet,
    NewSetting,
    #comm_class_exampleSource.ScriptExampleSet,
    #comm_class_exampleSource.ScriptSourceSet,
    #comm_class_exampleSource.ScriptDatasetSet,
    #comm_class_exampleSource.ExampleSet,
    comm_class_exampleSource.ExampleSourceSet,
    #comm_class_exampleSource.DatasetSet,
    #MyModel,
    #ReuseDataset,
                    comm_class.OutputSet,
                    comm_class.ConfigSet,
                    BaseModel):
    ...  

def FDStressModelRunner(opts:FDStressModel) -> int:
    try:
        logs = comm_class.myLog()

        paths = comm_func.create_path(str(opts.IO_output_path))
        output_path = paths["output_path"]
        work_path = paths["work_path"]
        download_path = paths["download_path"]

        logs.iprint("read source setting ...")
        datas = comm_class_exampleSource.read_source(opts,work_path,download_path,logs.iprint)

        pwd = os.getcwd()
        # prepare the fdforce
        os.chdir(work_path)
        allfiles = datas["all_files"]
        allexamples = [i for i in allfiles if os.path.isdir(i)]
        if len(allexamples) == 0:
            logs.iprint("no example found, exit! Please prepare the inputs of each example in each folder.")
            return 1
        if len(opts.stress_comp) == 0:
            logs.iprint("no stress component found, exit! Please specify the stress components.")
            return 1
        subfolders = FDStress.PrepareFDStress(allexamples,opts.fd_step,opts.fd_number,list(opts.stress_comp)).run()
        if len(subfolders) == 0:
            logs.iprint("no example found, exit! Please prepare the inputs of each example in each folder.")
            return 1

        allparams = {
            "config": comm_func.read_config(opts),
            "save_path": ".",
            "bohrium_group_name": "FDStress",
            "run_dft": {
                "example": subfolders,
                "command": opts.abacus_command,
                "image": opts.abacus_image,
                "bohrium": {
                    "scass_type": opts.bohrium_machine,
                    "job_type": "container",
                    "platform": "ali",
                },
            },
        }
        os.chdir(pwd)

        # execut
        stdout,stderr = comm_func.exec_abacustest(allparams,work_path)
        os.chdir(work_path)
        FDStress.PostProcessFDStress(allexamples,False).run()
        os.chdir(pwd)
        

        comm_report.gen_report(opts,logs,work_path,output_path,allparams)
    except:
        traceback.print_exc()
        return 1

    return 0
